#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from copy import deepcopy
import numpy as np
from numpy import ndarray


def range_re_seg(vol: np.ndarray,
                 roi: np.ndarray,
                 im_range=None) -> ndarray:
    """Removes voxels from the intensity mask that fall outside
    the given range (intensities outside the range are set to 0).

    Args:
        vol (ndarray): Imaging data.
        roi (ndarray): ROI mask with values of 0's and 1's.
        im_range (ndarray): 1-D array with shape (1,2) of the re-segmentation intensity range.

    Returns:
        ndarray: Intensity mask with intensities within the re-segmentation range.
    """

    if im_range is not None and len(im_range) == 2:
        roi = deepcopy(roi)
        roi[vol < im_range[0]] = 0
        roi[vol > im_range[1]] = 0

    return roi

def outlier_re_seg(vol: np.ndarray,
                   roi: np.ndarray,
                   outliers="") -> np.ndarray:
    """Removes voxels with outlier intensities from the given mask
    using the Collewet method.

    Args:
        vol (ndarray): Imaging data.
        roi (ndarray): ROI mask with values of 0 and 1.
        outliers (str, optional): Algo used to define outliers.
                                  (For now this methods only implements "Collewet" method).

    Returns:
        ndarray: An array with values of 0 and 1.

    Raises:
        ValueError: If `outliers` is not "Collewet" or None.

    Todo:
        * Delete outliers argument or implements others outlining methods.
    """

    if outliers != '':
        roi = deepcopy(roi)

        if outliers == "Collewet":
            u = np.mean(vol[roi == 1])
            sigma = np.std(vol[roi == 1])

            roi[vol > (u + 3*sigma)] = 0
            roi[vol < (u - 3*sigma)] = 0
        else:
            raise ValueError("Outlier segmentation not defined.")

    return roi



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List

import numpy as np
import pandas as pd


def rot_matrix(theta: float,
               dim: int=2,
               rot_axis: int=-1) -> np.ndarray:
    """Creates a 2d or 3d rotation matrix

    Args:
        theta (float): angle in radian
        dim (int, optional): dimension size. Defaults to 2.
        rot_axis (int, optional): rotation axis value. Defaults to -1.

    Returns:
        ndarray: rotation matrix
    """

    if dim == 2:
        rot_mat = np.array([[np.cos(theta), -np.sin(theta)],
                            [np.sin(theta), np.cos(theta)]])

    elif dim == 3:
        if rot_axis == 0:
            rot_mat = np.array([[1.0, 0.0, 0.0],
                                [0.0, np.cos(theta), -np.sin(theta)],
                                [0.0, np.sin(theta), np.cos(theta)]])
        elif rot_axis == 1:
            rot_mat = np.array([[np.cos(theta), 0.0, np.sin(theta)],
                                [0.0, 1.0, 0.0],
                                [-np.sin(theta), 0.0, np.cos(theta)]])
        elif rot_axis == 2:
            rot_mat = np.array([[np.cos(theta), -np.sin(theta), 0.0],
                                [np.sin(theta), np.cos(theta), 0.0],
                                [0.0, 0.0, 1.0]])
        else:
            rot_mat = None
    else:
        rot_mat = None

    return rot_mat


def sig_proc_segmentise(x: List) -> List:
    """Produces a list of segments from input x with values (0,1)

    Args:
        x (List): list of values

    Returns:
        List: list of segments from input x with values (0,1)
    """

    # Create a difference vector
    y = np.diff(x)

    # Find start and end indices of sections with value 1
    ind_1_start = np.array(np.where(y == 1)).flatten()
    if np.shape(ind_1_start)[0] > 0:
        ind_1_start += 1
    ind_1_end = np.array(np.where(y == -1)).flatten()

    # Check for boundary effects
    if x[0] == 1:
        ind_1_start = np.insert(ind_1_start, 0, 0)
    if x[-1] == 1:
        ind_1_end = np.append(ind_1_end, np.shape(x)[0]-1)

    # Generate segment df for segments with value 1
    if np.shape(ind_1_start)[0] == 0:
        df_one = pd.DataFrame({"i":   [],
                               "j":   [],
                               "val": []})
    else:
        df_one = pd.DataFrame({"i":   ind_1_start,
                               "j":   ind_1_end,
                               "val": np.ones(np.shape(ind_1_start)[0])})

    # Find start and end indices for section with value 0
    if np.shape(ind_1_start)[0] == 0:
        ind_0_start = np.array([0])
        ind_0_end = np.array([np.shape(x)[0]-1])
    else:
        ind_0_end = ind_1_start - 1
        ind_0_start = ind_1_end + 1

        # Check for boundary effect
        if x[0] == 0:
            ind_0_start = np.insert(ind_0_start, 0, 0)
        if x[-1] == 0:
            ind_0_end = np.append(ind_0_end, np.shape(x)[0]-1)

        # Check for out-of-range boundary effects
        if ind_0_end[0] < 0:
            ind_0_end = np.delete(ind_0_end, 0)
        if ind_0_start[-1] >= np.shape(x)[0]:
            ind_0_start = np.delete(ind_0_start, -1)

    # Generate segment df for segments with value 0
    if np.shape(ind_0_start)[0] == 0:
        df_zero = pd.DataFrame({"i":   [],
                                "j":   [],
                                "val": []})
    else:
        df_zero = pd.DataFrame({"i":    ind_0_start,
                                "j":    ind_0_end,
                                "val":  np.zeros(np.shape(ind_0_start)[0])})

    df_segm = df_one.append(df_zero).sort_values(by="i").reset_index(drop=True)

    return df_segm


def sig_proc_find_peaks(x: float,
                        ddir: str="pos") -> pd.DataFrame:
    """Determines peak positions in array of values

    Args:
        x (float): value
        ddir (str, optional): positive or negative value. Defaults to "pos".

    Returns:
        pd.DataFrame: peak positions in array of values
    """

    # Invert when looking for local minima
    if ddir == "neg":
        x = -x

    # Generate segments where slope is negative

    df_segm = sig_proc_segmentise(x=(np.diff(x) < 0.0)*1)

    # Start of slope coincides with position of peak (due to index shift induced by np.diff)
    ind_peak = df_segm.loc[df_segm.val == 1, "i"].values

    # Check right boundary
    if x[-1] > x[-2]:
        ind_peak = np.append(ind_peak, np.shape(x)[0]-1)

    # Construct dataframe with index and corresponding value
    if np.shape(ind_peak)[0] == 0:
        df_peak = pd.DataFrame({"ind": [],
                                "val": []})
    else:
        if ddir == "pos":
            df_peak = pd.DataFrame({"ind": ind_peak,
                                    "val": x[ind_peak]})
        if ddir == "neg":
            df_peak = pd.DataFrame({"ind":  ind_peak,
                                    "val": -x[ind_peak]})
    return df_peak



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Tuple, Union

import numpy as np


def mode(x: np.ndarray,
        return_counts=False) -> Union[Tuple[np.ndarray, np.ndarray],
                                      np.ndarray]:
    """Implementation of mode that also returns counts, unlike the standard statistics.mode.

    Args:
        x (ndarray): n-dimensional array of which to find mode.
        return_counts (bool): If True, also return the number of times each unique item appears in ``x``.

    Returns:
        2-element tuple containing
        
        - ndarray: Array of the modal (most common) value in the given array.
        - ndarray: Array of the counts if ``return_counts`` is True.
    """

    unique_values, counts = np.unique(x, return_counts=True)

    if return_counts:
        return unique_values[np.argmax(counts)], np.max(counts)

    return unique_values[np.argmax(counts)]



import json
import logging
import os
import pickle
import re
from dataclasses import dataclass
from pathlib import Path
from time import time
from typing import List, Union

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np
import pandas as pd
import pydicom
import pydicom.errors
import pydicom.misc
import ray
from nilearn import image
from numpyencoder import NumpyEncoder
from tqdm import tqdm, trange

from ..MEDscan import MEDscan
from ..processing.compute_suv_map import compute_suv_map
from ..processing.segmentation import get_roi_from_indexes
from ..utils.get_file_paths import get_file_paths
from ..utils.get_patient_names import get_patient_names
from ..utils.imref import imref3d
from ..utils.json_utils import load_json, save_json
from ..utils.save_MEDscan import save_MEDscan
from .ProcessDICOM import ProcessDICOM


class DataManager(object):
    """Reads all the raw data (DICOM, NIfTI) content and organizes it in instances of the MEDscan class."""


    @dataclass
    class DICOM(object):
        """DICOM data management class that will organize data during the conversion to MEDscan class process"""
        stack_series_rs: List
        stack_path_rs: List
        stack_frame_rs: List
        cell_series_id: List
        cell_path_rs: List
        cell_path_images: List
        cell_frame_rs: List
        cell_frame_id: List


    @dataclass
    class NIfTI(object):
        """NIfTI data management class that will organize data during the conversion to MEDscan class process"""
        stack_path_images: List
        stack_path_roi: List
        stack_path_all: List


    @dataclass
    class Paths(object):
        """Paths management class that will organize the paths used in the processing"""
        _path_to_dicoms: List
        _path_to_niftis: List
        _path_csv: Union[Path, str]
        _path_save: Union[Path, str]
        _path_save_checks: Union[Path, str]
        _path_pre_checks_settings: Union[Path, str]

    def __init__(
            self, 
            path_to_dicoms: List = [],
            path_to_niftis: List = [],
            path_csv: Union[Path, str] = None,
            path_save: Union[Path, str] = None,
            path_save_checks: Union[Path, str] = None,
            path_pre_checks_settings: Union[Path, str] = None,
            save: bool = True,
            n_batch: int = 2
    ) -> None:
        """Constructor of the class DataManager.

        Args:
            path_to_dicoms (Union[Path, str], optional): Full path to the starting directory
                where the DICOM data is located.
            path_to_niftis (Union[Path, str], optional): Full path to the starting directory
                where the NIfTI is located.
            path_csv (Union[Path, str], optional): Full path to the CSV file containing the scans info list.
            path_save (Union[Path, str], optional): Full path to the directory where to save all the MEDscan classes.
            path_save_checks(Union[Path, str], optional): Full path to the directory where to save all 
                the pre-radiomics checks analysis results.
            path_pre_checks_settings(Union[Path, str], optional): Full path to the JSON file of the pre-checks analysis
                parameters.
            save (bool, optional): True to save the MEDscan classes in `path_save`.
            n_batch (int, optional): Numerical value specifying the number of batch to use in the
                parallel computations (use 0 for serial computation).
        
        Returns:
            None
        """
        # Convert all paths to Pathlib.Path
        if path_to_dicoms:
            path_to_dicoms = Path(path_to_dicoms)
        if path_to_niftis:
            path_to_niftis = Path(path_to_niftis)
        if path_csv:
            path_csv = Path(path_csv)
        if path_save:
            path_save = Path(path_save)
        if path_save_checks:
            path_save_checks = Path(path_save_checks)
        if path_pre_checks_settings:
            path_pre_checks_settings = Path(path_pre_checks_settings)

        self.paths = self.Paths(
                path_to_dicoms,
                path_to_niftis,
                path_csv,
                path_save,
                path_save_checks,
                path_pre_checks_settings,
        )
        self.save = save
        self.n_batch = n_batch
        self.__dicom = self.DICOM(
                stack_series_rs=[],
                stack_path_rs=[],
                stack_frame_rs=[],
                cell_series_id=[],
                cell_path_rs=[],
                cell_path_images=[],
                cell_frame_rs=[],
                cell_frame_id=[]
        )
        self.__nifti = self.NIfTI(
                stack_path_images=[],
                stack_path_roi=[],
                stack_path_all=[]
        )
        self.path_to_objects = []
        self.summary = {}
        self.csv_data = None
        self.__studies = []
        self.__institutions = []
        self.__scans = []

    def __find_uid_cell_index(self, uid: Union[str, List[str]], cell: List[str]) -> List: 
        """Finds the cell with the same `uid`. If not is present in `cell`, creates a new position
        in the `cell` for the new `uid`.

        Args:
            uid (Union[str, List[str]]):  Unique identifier of the Series to find.
            cell (List[str]): List of Unique identifiers of the Series.

        Returns:
            Union[List[str], str]: List or string of the uid  
        """
        return [len(cell)] if uid not in cell else[i for i, e in enumerate(cell) if e == uid]

    def __get_list_of_files(self, dir_name: str) -> List:
        """Gets all files in the given directory

        Args:
            dir_name (str): directory name

        Returns:
            List: List of all files in the directory
        """
        list_of_file = os.listdir(dir_name)
        all_files = list()
        for entry in list_of_file:
            full_path = os.path.join(dir_name, entry)
            if os.path.isdir(full_path):
                all_files = all_files + self.__get_list_of_files(full_path)
            else:
                all_files.append(full_path)

        return all_files

    def __get_MEDscan_name_save(self, medscan: MEDscan) -> str:
        """Returns the name that will be used to save the MEDscan instance, based on the values of the attributes.

        Args:
            medscan(MEDscan): A MEDscan class instance.
        
        Returns:
            str: String of the name save.
        """
        series_description = medscan.series_description.translate({ord(ch): '-' for ch in '/\\ ()&:*'})
        name_id = medscan.patientID.translate({ord(ch): '-' for ch in '/\\ ()&:*'})
        # final saving name
        name_complete = name_id + '__' + series_description + '.' + medscan.type + '.npy'
        return name_complete

    def __associate_rt_stuct(self) -> None:
        """Associates the imaging volumes to their mask using UIDs

        Returns:
            None
        """
        print('--> Associating all RT objects to imaging volumes')
        n_rs = len(self.__dicom.stack_path_rs)
        self.__dicom.stack_series_rs = list(dict.fromkeys(self.__dicom.stack_series_rs))
        if n_rs:
            for i in trange(0, n_rs):
                try: 
                    # PUT ALL THE DICOM PATHS WITH THE SAME UID IN THE SAME PATH LIST
                    ind_series_id = self.__find_uid_cell_index(
                                                        self.__dicom.stack_series_rs[i], 
                                                        self.__dicom.cell_series_id)
                    for n in range(len(ind_series_id)):
                        if ind_series_id[n] < len(self.__dicom.cell_path_rs):
                            self.__dicom.cell_path_rs[ind_series_id[n]] += [self.__dicom.stack_path_rs[i]]
                except:
                    ind_series_id = self.__find_uid_cell_index(
                                                        self.__dicom.stack_frame_rs[i], 
                                                        self.__dicom.cell_frame_id)
                    for n in range(len(ind_series_id)):
                        if ind_series_id[n] < len(self.__dicom.cell_path_rs):
                            self.__dicom.cell_path_rs[ind_series_id[n]] += [self.__dicom.stack_path_rs[i]]
        print('DONE')

    def __read_all_dicoms(self) -> None:
        """Reads all the dicom files in the all the paths of the attribute `_path_to_dicoms`

        Returns:
            None
        """
        # SCANNING ALL FOLDERS IN INITIAL DIRECTORY
        print('\n--> Scanning all folders in initial directory...', end='')
        p = Path(self.paths._path_to_dicoms)
        e_rglob = '*.dcm'

        # EXTRACT ALL FILES IN THE PATH TO DICOMS
        if self.paths._path_to_dicoms.is_dir():
            stack_folder_temp = list(p.rglob(e_rglob))
            stack_folder = [x for x in stack_folder_temp if not x.is_dir()]
        elif str(self.paths._path_to_dicoms).find('json') != -1:
            with open(self.paths._path_to_dicoms) as f:
                data = json.load(f)
                for value in data.values():
                    stack_folder_temp = value
            directory_name = str(stack_folder_temp).replace("'", '').replace('[', '').replace(']', '')
            stack_folder = self.__get_list_of_files(directory_name)
        else:
            raise ValueError("The given dicom folder path either doesn't exist or not a folder.")
        # READ ALL DICOM FILES AND UPDATE ATTRIBUTES FOR FURTHER PROCESSING
        for file in tqdm(stack_folder):
            if pydicom.misc.is_dicom(file):
                try:
                    info = pydicom.dcmread(str(file))
                    if info.Modality in ['MR', 'PT', 'CT']:
                        ind_series_id = self.__find_uid_cell_index(
                                                        info.SeriesInstanceUID, 
                                                        self.__dicom.cell_series_id)[0]
                        if ind_series_id == len(self.__dicom.cell_series_id):  # New volume
                            self.__dicom.cell_series_id = self.__dicom.cell_series_id + [info.SeriesInstanceUID]
                            self.__dicom.cell_frame_id += [info.FrameOfReferenceUID]
                            self.__dicom.cell_path_images += [[]]
                            self.__dicom.cell_path_rs = self.__dicom.cell_path_rs + [[]]
                        self.__dicom.cell_path_images[ind_series_id] += [file]
                    elif info.Modality == 'RTSTRUCT':
                        self.__dicom.stack_path_rs += [file]
                        try:
                            series_uid = info.ReferencedFrameOfReferenceSequence[
                                        0].RTReferencedStudySequence[
                                        0].RTReferencedSeriesSequence[
                                        0].SeriesInstanceUID
                        except:
                            series_uid = 'NotFound'
                        self.__dicom.stack_series_rs += [series_uid]
                        try:
                            frame_uid = info.ReferencedFrameOfReferenceSequence[0].FrameOfReferenceUID
                        except:
                            frame_uid = info.FrameOfReferenceUID
                        self.__dicom.stack_frame_rs += [frame_uid]
                    else:
                        print("Modality not supported: ", info.Modality)

                except Exception as e:
                    print(f'Error while reading: {file}, error: {e}\n')
                    continue
        print('DONE')

        # ASSOCIATE ALL VOLUMES TO THEIR MASK
        self.__associate_rt_stuct()

    def process_all_dicoms(self) -> Union[List[MEDscan], None]:
        """This function reads the DICOM content of all the sub-folder tree of a starting directory defined by
        `path_to_dicoms`. It then organizes the data (files throughout the starting directory are associated by
        'SeriesInstanceUID') in the MEDscan class including the region of  interest (ROI) defined by an
        associated RTstruct. All MEDscan classes hereby created are saved in `path_save` with a name
        varying with every scan.

        Returns:
            List[MEDscan]: List of MEDscan instances.
        """
        ray.init(local_mode=True, include_dashboard=True)

        print('--> Reading all DICOM objects to create MEDscan classes')
        self.__read_all_dicoms()

        print('--> Processing DICOMs and creating MEDscan objects')
        n_scans = len(self.__dicom.cell_path_images)
        if self.n_batch is None:
            n_batch = 1
        elif n_scans < self.n_batch:
            n_batch = n_scans
        else:
            n_batch = self.n_batch

        # Distribute the first tasks to all workers
        pds = [ProcessDICOM(
                        self.__dicom.cell_path_images[i], 
                        self.__dicom.cell_path_rs[i], 
                        self.paths._path_save,
                        self.save)
            for i in range(n_batch)]
        
        ids = [pd.process_files() for pd in pds]

        # Update the path to the created instances
        for name_save in ray.get(ids):
            if self.paths._path_save:
                self.path_to_objects.append(str(self.paths._path_save / name_save))
            # Update processing summary
            if name_save.split('_')[0].count('-') >= 2:
                scan_type = name_save[name_save.find('__')+2 : name_save.find('.')]
                if name_save.split('-')[0] not in self.__studies:
                    self.__studies.append(name_save.split('-')[0])  # add new study
                if name_save.split('-')[1] not in self.__institutions:
                    self.__institutions.append(name_save.split('-')[1])  # add new study
                if name_save.split('-')[0] not in self.summary:
                    self.summary[name_save.split('-')[0]] = {}
                if name_save.split('-')[1] not  in self.summary[name_save.split('-')[0]]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]] = {}  # add new institution
                if scan_type not in self.__scans:
                    self.__scans.append(scan_type)
                if scan_type not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type] = []
                if name_save not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type].append(name_save)
            else:
                if self.save:
                    logging.warning(f"The patient ID of the following file: {name_save} does not respect the MEDiml "\
                        "naming convention 'study-institution-id' (Ex: Glioma-TCGA-001)")

        nb_job_left = n_scans - n_batch

        # Distribute the remaining tasks
        for _ in trange(n_scans):
            _, ids = ray.wait(ids, num_returns=1)
            if nb_job_left > 0:
                idx = n_scans - nb_job_left
                pd = ProcessDICOM(
                        self.__dicom.cell_path_images[idx], 
                        self.__dicom.cell_path_rs[idx], 
                        self.paths._path_save,
                        self.save)
                ids.extend([pd.process_files()])
                nb_job_left -= 1

            # Update the path to the created instances
            for name_save in ray.get(ids):
                if self.paths._path_save:
                    self.path_to_objects.extend(str(self.paths._path_save / name_save))
                # Update processing summary
                if name_save.split('_')[0].count('-') >= 2:
                    scan_type = name_save[name_save.find('__')+2 : name_save.find('.')]
                    if name_save.split('-')[0] not in self.__studies:
                        self.__studies.append(name_save.split('-')[0])  # add new study
                    if name_save.split('-')[1] not in self.__institutions:
                        self.__institutions.append(name_save.split('-')[1])  # add new study
                    if name_save.split('-')[0] not in self.summary:
                        self.summary[name_save.split('-')[0]] = {}
                    if name_save.split('-')[1] not  in self.summary[name_save.split('-')[0]]:
                        self.summary[name_save.split('-')[0]][name_save.split('-')[1]] = {}  # add new institution
                    if scan_type not in self.__scans:
                        self.__scans.append(scan_type)
                    if scan_type not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]]:
                        self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type] = []
                    if name_save not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type]:
                        self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type].append(name_save)
                else:
                    if self.save:
                        logging.warning(f"The patient ID of the following file: {name_save} does not respect the MEDiml "\
                            "naming convention 'study-institution-id' (Ex: Glioma-TCGA-001)")
        print('DONE')

    def __read_all_niftis(self) -> None:
        """Reads all files in the initial path and organizes other path to images and roi
        in the class attributes.

        Returns:
            None.
        """
        print('\n--> Scanning all folders in initial directory')
        if not self.paths._path_to_niftis:
            raise ValueError("The path to the niftis is not defined")
        p = Path(self.paths._path_to_niftis)
        e_rglob1 = '*.nii'
        e_rglob2 = '*.nii.gz'

        # EXTRACT ALL FILES IN THE PATH TO DICOMS
        if p.is_dir():
            self.__nifti.stack_path_all = list(p.rglob(e_rglob1))
            self.__nifti.stack_path_all.extend(list(p.rglob(e_rglob2)))
        else:
            raise TypeError(f"{p} must be a path to a directory")

        all_niftis = list(self.__nifti.stack_path_all)
        for i in trange(0, len(all_niftis)):
            if 'ROI' in all_niftis[i].name.split("."):
                self.__nifti.stack_path_roi.append(all_niftis[i])
            else:
                self.__nifti.stack_path_images.append(all_niftis[i])
        print('DONE')

    def __associate_roi_to_image(
            self,
            image_file: Union[Path, str], 
            medscan: MEDscan, 
            nifti: nib.Nifti1Image,
            path_roi_data: Path = None
        ) -> MEDscan:
        """Extracts all ROI data from the given path for the given patient ID and updates all class attributes with
        the new extracted data.

        Args:
            image_file(Union[Path, str]): Path to the ROI data.
            medscan (MEDscan): MEDscan class instance that will hold the data. 

        Returns:
            MEDscan: Returns a MEDscan instance with updated roi attributes.
        """
        def load_mask(_id, file, medscan):
            # Load the patient's ROI nifti files:
            if file.name.startswith(_id) and 'ROI' in file.name.split("."):
                roi = nib.load(file)
                roi = image.resample_to_img(roi, nifti, interpolation='nearest')
                roi_data = roi.get_fdata()
                roi_name = file.name[file.name.find("(") + 1 : file.name.find(")")]
                name_set = file.name[file.name.find("_") + 2 : file.name.find("(")]
                medscan.data.ROI.update_indexes(key=roi_index, indexes=np.nonzero(roi_data.flatten()))
                medscan.data.ROI.update_name_set(key=roi_index, name_set=name_set)
                medscan.data.ROI.update_roi_name(key=roi_index, roi_name=roi_name)  
            else:
                raise ValueError(f"The ROI file for patient ID: {_id} "
                    f"was not found in the given path: {file} or was not correctly named.")   
        
        image_file = Path(image_file)
        roi_index = 0
        if not path_roi_data:
            if not self.paths._path_to_niftis:
                raise ValueError("The path to the niftis is not defined")
            else:
                path_roi_data = self.paths._path_to_niftis

            for file in self.__nifti.stack_path_roi:
                _id = image_file.name.split("(")[0] if ("(") in image_file.name else image_file.name # id is PatientID__ImagingScanName
                load_mask(_id, file, medscan)
                roi_index += 1
        else:
            _id = image_file.name.split("(")[0] if ("(") in image_file.name else image_file.name # id is PatientID__ImagingScanName
            load_mask(_id, path_roi_data, medscan)

        return medscan

    def __associate_spatialRef(self, nifti_file: Union[Path, str], medscan: MEDscan) -> MEDscan:
        """Computes the imref3d spatialRef using a NIFTI file and updates the spatialRef attribute.

        Args:
            nifti_file(Union[Path, str]): Path to the nifti data.
            medscan (MEDscan): MEDscan class instance that will hold the data.

        Returns:
            MEDscan: Returns a MEDscan instance with updated spatialRef attribute.
        """
        # Loading the nifti file :
        nifti = nib.load(nifti_file)
        nifti_data = medscan.data.volume.array

        # spatialRef Creation
        pixel_x = abs(nifti.affine[0, 0])
        pixel_y = abs(nifti.affine[1, 1])
        slices = abs(nifti.affine[2, 2])
        min_grid = nifti.affine[:3, 3] * [-1.0, -1.0, 1.0] # x and y are flipped
        min_x_grid = min_grid[0]
        min_y_grid = min_grid[1]
        min_z_grid = min_grid[2]
        size_image = np.shape(nifti_data)
        spatialRef = imref3d(size_image, abs(pixel_x), abs(pixel_y), abs(slices))
        spatialRef.XWorldLimits = (np.array(spatialRef.XWorldLimits) -
                                (spatialRef.XWorldLimits[0] -
                                    (min_x_grid-pixel_x/2))
                                ).tolist()
        spatialRef.YWorldLimits = (np.array(spatialRef.YWorldLimits) -
                                (spatialRef.YWorldLimits[0] -
                                    (min_y_grid-pixel_y/2))
                                ).tolist()
        spatialRef.ZWorldLimits = (np.array(spatialRef.ZWorldLimits) -
                                (spatialRef.ZWorldLimits[0] -
                                    (min_z_grid-slices/2))
                                ).tolist()

        # Converting the results into lists
        spatialRef.ImageSize = spatialRef.ImageSize.tolist()
        spatialRef.XIntrinsicLimits = spatialRef.XIntrinsicLimits.tolist()
        spatialRef.YIntrinsicLimits = spatialRef.YIntrinsicLimits.tolist()
        spatialRef.ZIntrinsicLimits = spatialRef.ZIntrinsicLimits.tolist()

        # update spatialRef in the volume sub-class
        medscan.data.volume.update_spatialRef(spatialRef)

        return medscan

    def __process_one_nifti(self, nifti_file: Union[Path, str], path_data) -> MEDscan:
        """
        Processes one NIfTI file to create a MEDscan class instance.

        Args:
            nifti_file (Union[Path, str]): Path to the NIfTI file.
            path_data (Union[Path, str]): Path to the data.
        
        Returns:
            MEDscan: MEDscan class instance.
        """
        medscan = MEDscan()
        medscan.patientID = os.path.basename(nifti_file).split("_")[0]
        medscan.type = os.path.basename(nifti_file).split(".")[-3]
        medscan.series_description = nifti_file.name[nifti_file.name.find('__') + 2: nifti_file.name.find('(')]
        medscan.format = "nifti"
        medscan.data.set_orientation(orientation="Axial")
        medscan.data.set_patient_position(patient_position="HFS")
        medscan.data.volume.array = nib.load(nifti_file).get_fdata()
        medscan.data.volume.scan_rot = None
        
        # Update spatialRef
        self.__associate_spatialRef(nifti_file, medscan)
        
        # Assiocate ROI
        medscan = self.__associate_roi_to_image(nifti_file, medscan, nib.load(nifti_file), path_data)

        return medscan
    
    def process_all(self) -> None:
        """Processes both DICOM & NIfTI content to create MEDscan classes
        """
        self.process_all_dicoms()
        self.process_all_niftis()

    def process_all_niftis(self) -> List[MEDscan]:
        """This function reads the NIfTI content of all the sub-folder tree of a starting directory. 
        It then organizes the data in the MEDscan class including the region of  interest (ROI)
        defined by an associated mask file. All MEDscan classes hereby created are saved in a specific path
        with a name specific name varying with every scan.

        Args:
            None.

        Returns:
            List[MEDscan]: List of MEDscan instances.
        """

        # Reading all NIfTI files
        self.__read_all_niftis()

        # Create the MEDscan instances
        print('--> Reading all NIfTI objects (imaging volumes & masks) to create MEDscan classes')
        list_instances = []
        for file in tqdm(self.__nifti.stack_path_images):
            # Assert the list of instances does not exceed the a size of 10
            if len(list_instances) >= 10:
                print('The number of MEDscan instances exceeds 10, please consider saving the instances')
                break
            # INITIALIZE MEDscan INSTANCE AND UPDATE ATTRIBUTES
            medscan = MEDscan()
            medscan.patientID = os.path.basename(file).split("_")[0]
            medscan.type = os.path.basename(file).split(".")[-3]
            medscan.series_description = file.name[file.name.find('__') + 2: file.name.find('(')] if '__' in file.name else ""
            medscan.format = "nifti"
            medscan.data.set_orientation(orientation="Axial")
            medscan.data.set_patient_position(patient_position="HFS")
            medscan.data.volume.array = nib.load(file).get_fdata()
            
            # RAS to LPS
            #medscan.data.volume.convert_to_LPS()
            medscan.data.volume.scan_rot = None
            
            # Update spatialRef
            medscan = self.__associate_spatialRef(file, medscan)
            
            # Get ROI
            medscan = self.__associate_roi_to_image(file, medscan, nib.load(file))

            # SAVE MEDscan INSTANCE
            if self.save and self.paths._path_save:
                save_MEDscan(medscan, self.paths._path_save)
            else:
                list_instances.append(medscan)
            
            # Update the path to the created instances
            name_save = self.__get_MEDscan_name_save(medscan)

            # Clear memory
            del medscan

            # Update the path to the created instances
            if self.paths._path_save:
                self.path_to_objects.append(str(self.paths._path_save / name_save))
            
            # Update processing summary
            if name_save.split('_')[0].count('-') >= 2:
                scan_type = name_save[name_save.find('__')+2 : name_save.find('.')]
                if name_save.split('-')[0] not in self.__studies:
                    self.__studies.append(name_save.split('-')[0])  # add new study
                if name_save.split('-')[1] not in self.__institutions:
                    self.__institutions.append(name_save.split('-')[1])  # add new institution
                if name_save.split('-')[0] not in self.summary:
                    self.summary[name_save.split('-')[0]] = {}  # add new study to summary
                if name_save.split('-')[1] not  in self.summary[name_save.split('-')[0]]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]] = {}  # add new institution
                if scan_type not in self.__scans:
                    self.__scans.append(scan_type)
                if scan_type not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type] = []
                if name_save not in self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type]:
                    self.summary[name_save.split('-')[0]][name_save.split('-')[1]][scan_type].append(name_save)
            else:
                if self.save:
                    logging.warning(f"The patient ID of the following file: {name_save} does not respect the MEDiml "\
                        "naming convention 'study-institution-id' (Ex: Glioma-TCGA-001)")
        print('DONE')

        if list_instances:
            return list_instances

    def process_one_nifti(self, path_image: Union[Path, str], path_mask: Union[Path, str]) -> MEDscan:
        """Processes one NIfTI file to create a MEDscan class instance.

        Args:
            nifti_file (Union[Path, str]): Path to the NIfTI file.
            path_data (Union[Path, str]): Path to the data.
        
        Returns:
            MEDscan: MEDscan class instance.
        """
        medscan = self.__process_one_nifti(path_image, path_mask)

        # SAVE MEDscan INSTANCE
        if self.save and self.paths._path_save:
            save_MEDscan(medscan, self.paths._path_save)

        return medscan
    
    def update_from_csv(self, path_csv: Union[str, Path] = None) -> None:
        """Updates the class from a given CSV and summarizes the processed scans again according to it.

        Args:
            path_csv(optional, Union[str, Path]): Path to a csv file, if not given, will check
                for csv info in the class attributes.
        
        Returns:
            None
        """
        if not (path_csv or self.paths._path_csv):
            print('No csv provided, no updates will be made')
        else:
            if path_csv:
                self.paths._path_csv = path_csv
            # Extract roi type label from csv file name
            name_csv = self.paths._path_csv.name
            roi_type_label = name_csv[name_csv.find('_')+1 : name_csv.find('.')]

            # Create a dictionary
            csv_data = {}
            csv_data[roi_type_label] = pd.read_csv(self.paths._path_csv)
            self.csv_data = csv_data
            self.summarize()

    def summarize(self, retrun_summary: bool = False) -> None:
        """Creates and shows a summary of processed scans organized by study, institution, scan type and roi type

        Args:
            retrun_summary (bool, optional): If True, will return the summary as a dictionary.
        
        Returns:
            None
        """
        def count_scans(summary):
            count = 0
            if type(summary) == dict:
                for study in summary:
                    if type(summary[study]) == dict:
                        for institution in summary[study]:
                            if type(summary[study][institution]) == dict:
                                for scan in self.summary[study][institution]:
                                    count += len(summary[study][institution][scan])
                            else:
                                count += len(summary[study][institution])
                    else:
                        count += len(summary[study])
            elif type(summary) == list:
                count = len(summary)
            return count

        summary_df = pd.DataFrame(columns=['study', 'institution', 'scan_type', 'roi_type', 'count'])

        for study in self.summary:
            summary_df = summary_df.append({
                                        'study': study,
                                        'institution': "",
                                        'scan_type': "",
                                        'roi_type': "",
                                        'count' : count_scans(self.summary)
                                        }, ignore_index=True)
            for institution in self.summary[study]:
                summary_df = summary_df.append({
                                            'study': study,
                                            'institution': institution,
                                            'scan_type': "",
                                            'roi_type': "",
                                            'count' : count_scans(self.summary[study][institution])
                                            }, ignore_index=True)
                for scan in self.summary[study][institution]:
                    summary_df = summary_df.append({
                                                'study': study,
                                                'institution': institution,
                                                'scan_type': scan,
                                                'roi_type': "",
                                                'count' : count_scans(self.summary[study][institution][scan])
                                                }, ignore_index=True)
                    if self.csv_data:
                        roi_count = 0
                        for roi_type in self.csv_data:
                            csv_table = pd.DataFrame(self.csv_data[roi_type])
                            csv_table['under'] = '_'
                            csv_table['dot'] = '.'
                            csv_table['npy'] = '.npy'
                            name_patients = (pd.Series(
                                csv_table[['PatientID', 'under', 'under',
                                        'ImagingScanName',
                                        'dot',
                                        'ImagingModality',
                                        'npy']].fillna('').values.tolist()).str.join('')).tolist()
                            for patient_id in self.summary[study][institution][scan]:
                                if patient_id in name_patients:
                                    roi_count += 1
                            summary_df = summary_df.append({
                                                'study': study,
                                                'institution': institution,
                                                'scan_type': scan,
                                                'roi_type': roi_type,
                                                'count' : roi_count
                                                }, ignore_index=True)
        print(summary_df.to_markdown(index=False))

        if retrun_summary:
            return summary_df

    def __pre_radiomics_checks_dimensions(
        self,
        path_data: Union[Path, str] = None,
        wildcards_dimensions: List[str] = [],
        min_percentile: float = 0.05,
        max_percentile: float = 0.95,
        save: bool = False
        ) -> None:
        """Finds proper voxels dimension options for radiomics analyses for a group of scans

        Args:
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            wildcards_dimensions(List[str], optional): List of wildcards that determines the scans 
                that will be analyzed. You can learn more about wildcards in
                :ref:`this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`.
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
            save (bool, optional): If True, will save the results in a json file. Defaults to False.
        
        Returns:
            None.
        """
        xy_dim = {
            "data": [],
            "mean": [],
            "median": [],
            "std": [],
            "min": [],
            "max": [],
            f"p{min_percentile}": [],
            f"p{max_percentile}": []
        }
        z_dim = {
            "data": [],
            "mean": [],
            "median": [],
            "std": [],
            "min": [],
            "max": [],
            f"p{min_percentile}": [],
            f"p{max_percentile}": []
        }
        if type(wildcards_dimensions) is str:
            wildcards_dimensions = [wildcards_dimensions]

        if len(wildcards_dimensions) == 0:
            print("Wildcard is empty, the pre-checks will be aborted")
            return

        # Updating plotting params
        plt.rcParams["figure.figsize"] = (20,20)
        plt.rcParams.update({'font.size': 22})

        # TODO: seperate by studies and scan type (MRscan, CTscan...)
        # TODO: Two summaries (df, list of names saves) -> 
        # name_save = name_save(ROI) : Glioma-Huashan-001__T1.MRscan.npy({GTV})
        file_paths = list()
        for w in range(len(wildcards_dimensions)):
            wildcard = wildcards_dimensions[w]
            if path_data:
                file_paths = get_file_paths(path_data, wildcard)
            elif self.paths._path_save:
                file_paths = get_file_paths(self.paths._path_save, wildcard)
            else:
                raise ValueError("Path data is invalid.")
            n_files = len(file_paths)
            xy_dim["data"] = np.zeros((n_files, 1))
            xy_dim["data"] = np.multiply(xy_dim["data"], np.nan)
            z_dim["data"] = np.zeros((n_files, 1))
            z_dim["data"] = np.multiply(z_dim["data"], np.nan)
            for f in tqdm(range(len(file_paths))):
                try:
                    if file_paths[f].name.endswith("nii.gz") or file_paths[f].name.endswith("nii"):
                        medscan = nib.load(file_paths[f])
                        xy_dim["data"][f] = medscan.header.get_zooms()[0]
                        z_dim["data"][f]  = medscan.header.get_zooms()[2]
                    else:
                        medscan = np.load(file_paths[f], allow_pickle=True)
                        xy_dim["data"][f] = medscan.data.volume.spatialRef.PixelExtentInWorldX
                        z_dim["data"][f]  = medscan.data.volume.spatialRef.PixelExtentInWorldZ
                except Exception as e:
                    print(e)

            # Running analysis
            xy_dim["data"] = np.concatenate(xy_dim["data"])
            xy_dim["mean"] = np.mean(xy_dim["data"][~np.isnan(xy_dim["data"])])
            xy_dim["median"] = np.median(xy_dim["data"][~np.isnan(xy_dim["data"])])
            xy_dim["std"] = np.std(xy_dim["data"][~np.isnan(xy_dim["data"])])
            xy_dim["min"] = np.min(xy_dim["data"][~np.isnan(xy_dim["data"])])
            xy_dim["max"] = np.max(xy_dim["data"][~np.isnan(xy_dim["data"])])
            xy_dim[f"p{min_percentile}"] = np.percentile(xy_dim["data"][~np.isnan(xy_dim["data"])], 
                                                        min_percentile)
            xy_dim[f"p{max_percentile}"] = np.percentile(xy_dim["data"][~np.isnan(xy_dim["data"])], 
                                                        max_percentile)
            z_dim["mean"] = np.mean(z_dim["data"][~np.isnan(z_dim["data"])])
            z_dim["median"] = np.median(z_dim["data"][~np.isnan(z_dim["data"])])
            z_dim["std"] = np.std(z_dim["data"][~np.isnan(z_dim["data"])])
            z_dim["min"] = np.min(z_dim["data"][~np.isnan(z_dim["data"])])
            z_dim["max"] = np.max(z_dim["data"][~np.isnan(z_dim["data"])])
            z_dim[f"p{min_percentile}"] = np.percentile(z_dim["data"][~np.isnan(z_dim["data"])], 
                                                        min_percentile)
            z_dim[f"p{max_percentile}"] = np.percentile(z_dim["data"][~np.isnan(z_dim["data"])], max_percentile)
            xy_dim["data"] = xy_dim["data"].tolist()
            z_dim["data"] = z_dim["data"].tolist()
            
            # Plotting xy-spacing data histogram
            df_xy = pd.DataFrame(xy_dim["data"], columns=['data'])
            del xy_dim["data"]  # no interest in keeping data (we only need statistics)
            ax = df_xy.hist(column='data')
            min_quant, max_quant, median = df_xy.quantile(min_percentile), df_xy.quantile(max_percentile), df_xy.median()
            for x in ax[0]:
                x.axvline(min_quant.data, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.data, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(median.data, linestyle='solid', color='gold', label=f"Median: {float(median.data):.3f}")
                x.grid(False)
                plt.title(f"Voxels xy-spacing checks for {wildcard}")
                plt.legend()
                # Save the plot
                if save:
                    plt.savefig(self.paths._path_save_checks / ('Voxels_xy_check.png'))
                else:
                    plt.show()
            
            # Plotting z-spacing data histogram
            df_z = pd.DataFrame(z_dim["data"], columns=['data'])
            del z_dim["data"]  # no interest in keeping data (we only need statistics)
            ax = df_z.hist(column='data')
            min_quant, max_quant, median = df_z.quantile(min_percentile), df_z.quantile(max_percentile), df_z.median()
            for x in ax[0]:
                x.axvline(min_quant.data, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.data, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(median.data, linestyle='solid', color='gold', label=f"Median: {float(median.data):.3f}")
                x.grid(False)
                plt.title(f"Voxels z-spacing checks for {wildcard}")
                plt.legend()
                # Save the plot
                if save:
                    plt.savefig(self.paths._path_save_checks / ('Voxels_z_check.png'))
                else:
                    plt.show()
                
            # Saving files using wildcard for name
            if save:
                wildcard = str(wildcard).replace('*', '').replace('.npy', '.json')
                save_json(self.paths._path_save_checks / ('xyDim_' + wildcard), xy_dim, cls=NumpyEncoder)
                save_json(self.paths._path_save_checks / ('zDim_' + wildcard), z_dim, cls=NumpyEncoder)

    def __pre_radiomics_checks_window(
        self,
        path_data: Union[str, Path] = None,
        wildcards_window: List = [], 
        path_csv: Union[str, Path] = None,
        min_percentile: float = 0.05,
        max_percentile: float = 0.95,
        bin_width: int = 0,
        hist_range: list = [],
        nifti: bool = True,
        save: bool = False
        ) -> None:
        """Finds proper re-segmentation ranges options for radiomics analyses for a group of scans

        Args:
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            wildcards_window(List[str], optional): List of wildcards that determines the scans 
                that will be analyzed. You can learn more about wildcards in
                :ref:`this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`.
            path_csv(Union[str, Path], optional): Path to a csv file containing a list of the scans that will be
                analyzed (a CSV file for a single ROI type).
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
            bin_width(int, optional): Width of the bins for the histograms. If not provided, will use the 
                default number of bins in the method 
                :ref:`pandas.DataFrame.hist <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.hist.html>`: 10 bins.
            hist_range(list, optional): Range of the histograms. If empty, will use the minimum and maximum values.
            nifti(bool, optional): If True, will use the NIfTI files, otherwise will use the numpy files.
            save (bool, optional): If True, will save the results in a json file. Defaults to False.
        
        Returns:
            None.
        """
        # Updating plotting params
        plt.rcParams["figure.figsize"] = (20,20)
        plt.rcParams.update({'font.size': 22})
        
        if type(wildcards_window) is str:
            wildcards_window = [wildcards_window]

        if len(wildcards_window) == 0:
            print("Wilcards is empty")
            return
        if path_csv:
            self.paths._path_csv = Path(path_csv)
        roi_table = pd.read_csv(self.paths._path_csv)
        if nifti:
            roi_table['under'] = '_'
            roi_table['dot'] = '.'
            roi_table['roi_label'] = 'GTV'
            roi_table['oparenthesis'] = '('
            roi_table['cparenthesis'] = ')'
            roi_table['ext'] = '.nii.gz'
            patient_names = (pd.Series(
                roi_table[['PatientID', 'under', 'under',
                        'ImagingScanName',
                        'oparenthesis',
                        'roi_label',
                        'cparenthesis',
                        'dot',
                        'ImagingModality',
                        'ext']].fillna('').values.tolist()).str.join('')).tolist()
        else:
            roi_names = [[], [], []]
            roi_names[0] = roi_table['PatientID']
            roi_names[1] = roi_table['ImagingScanName']
            roi_names[2] = roi_table['ImagingModality']
            patient_names = get_patient_names(roi_names)
        for w in range(len(wildcards_window)):
            temp_val = []
            temp = []
            file_paths = []
            roi_data= {
                "data": [],
                "mean": [],
                "median": [],
                "std": [],
                "min": [],
                "max": [],
                f"p{min_percentile}": [],
                f"p{max_percentile}": []
            }
            wildcard = wildcards_window[w]
            if path_data:
                file_paths = get_file_paths(path_data, wildcard)
            elif self.paths._path_save:
                path_data = self.paths._path_save
                file_paths = get_file_paths(self.paths._path_save, wildcard)
            else:
                raise ValueError("Path data is invalid.")
            n_files = len(file_paths)
            i = 0
            for f in tqdm(range(n_files)):
                file = file_paths[f]
                _, filename = os.path.split(file)
                filename, ext = os.path.splitext(filename)
                patient_name = filename + ext
                try:
                    if file.name.endswith('nii.gz') or file.name.endswith('nii'):
                        medscan = self.__process_one_nifti(file, path_data)
                    else:
                        medscan = np.load(file, allow_pickle=True)
                        if re.search('PTscan', wildcard) and medscan.format != 'nifti':
                            medscan.data.volume.array = compute_suv_map(
                                                        np.double(medscan.data.volume.array), 
                                                        medscan.dicomH[2])
                    patient_names = pd.Index(patient_names)
                    ind_roi = patient_names.get_loc(patient_name)
                    name_roi = roi_table.loc[ind_roi][3]
                    vol_obj_init, roi_obj_init = get_roi_from_indexes(medscan, name_roi, 'box')
                    temp = vol_obj_init.data[roi_obj_init.data == 1]
                    temp_val.append(len(temp))
                    roi_data["data"].append(np.zeros(shape=(n_files, temp_val[i])))
                    roi_data["data"][i] = temp
                    i+=1
                    del medscan
                    del vol_obj_init
                    del roi_obj_init
                except Exception as e:
                    print(f"Problem with patient {patient_name}, error: {e}")
            
            roi_data["data"] = np.concatenate(roi_data["data"])
            roi_data["mean"] = np.mean(roi_data["data"][~np.isnan(roi_data["data"])])
            roi_data["median"] = np.median(roi_data["data"][~np.isnan(roi_data["data"])])
            roi_data["std"] = np.std(roi_data["data"][~np.isnan(roi_data["data"])])
            roi_data["min"] = np.min(roi_data["data"][~np.isnan(roi_data["data"])])
            roi_data["max"] = np.max(roi_data["data"][~np.isnan(roi_data["data"])])
            roi_data[f"p{min_percentile}"] = np.percentile(roi_data["data"][~np.isnan(roi_data["data"])], 
                                                        min_percentile)
            roi_data[f"p{max_percentile}"] = np.percentile(roi_data["data"][~np.isnan(roi_data["data"])], 
                                                        max_percentile)
            
            # Set bin width if not provided
            if bin_width != 0:
                if hist_range:
                    nb_bins = (round(hist_range[1]) - round(hist_range[0])) // bin_width
                else:
                    nb_bins = (round(roi_data["max"]) - round(roi_data["min"])) // bin_width
            else:
                nb_bins = 10
                if hist_range:
                    bin_width = int((round(hist_range[1]) - round(hist_range[0])) // nb_bins)
                else:
                    bin_width = int((round(roi_data["max"]) - round(roi_data["min"])) // nb_bins)
            nb_bins = int(nb_bins)

            # Set histogram range if not provided
            if not hist_range:
                hist_range = (roi_data["min"], roi_data["max"])

           # re-segment data according to histogram range
            roi_data["data"] = roi_data["data"][(roi_data["data"] > hist_range[0]) & (roi_data["data"] < hist_range[1])]
            df_data = pd.DataFrame(roi_data["data"], columns=['data'])
            del roi_data["data"]  # no interest in keeping data (we only need statistics)

            # Plot histogram
            ax = df_data.hist(column='data', bins=nb_bins, range=(hist_range[0], hist_range[1]), edgecolor='black')
            min_quant, max_quant= df_data.quantile(min_percentile), df_data.quantile(max_percentile)
            for x in ax[0]:
                x.axvline(min_quant.data, linestyle=':', color='r', label=f"{min_percentile*100}% Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.data, linestyle=':', color='g', label=f"{max_percentile*100}% Percentile: {float(max_quant):.3f}")
                x.grid(False)
                x.xaxis.set_ticks(np.arange(hist_range[0], hist_range[1], bin_width, dtype=int))
                x.set_xticklabels(x.get_xticks(), rotation=45)
                x.xaxis.set_tick_params(pad=15)
                plt.title(f"Intensity range checks for {wildcard}, bw={bin_width}")
                plt.legend()
                # Save the plot
                if save:
                    plt.savefig(self.paths._path_save_checks / ('Intensity_range_check_' + f'bw_{bin_width}.png'))
                else:
                    plt.show()
            
            # save final checks
            if save:
                wildcard = str(wildcard).replace('*', '').replace('.npy', '.json')
                save_json(self.paths._path_save_checks / ('roi_data_' + wildcard), roi_data, cls=NumpyEncoder)

    def pre_radiomics_checks(self,
                            path_data: Union[str, Path] = None,
                            wildcards_dimensions: List = [],
                            wildcards_window: List = [],
                            path_csv: Union[str, Path] = None,
                            min_percentile: float = 0.05,
                            max_percentile: float = 0.95,
                            bin_width: int = 0,
                            hist_range: list = [],
                            nifti: bool = False,
                            save: bool = False) -> None:
        """Finds proper dimension and re-segmentation ranges options for radiomics analyses. 

        The resulting files from this method can then be analyzed and used to set up radiomics 
        parameters options in computation methods.

        Args:
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            wildcards_dimensions(List[str], optional): List of wildcards that determines the scans 
                that will be analyzed. You can learn more about wildcards in
                `this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`_.
            wildcards_window(List[str], optional): List of wildcards that determines the scans 
                that will be analyzed. You can learn more about wildcards in
                `this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`_.
            path_csv(Union[str, Path], optional): Path to a csv file containing a list of the scans that will be
                analyzed (a CSV file for a single ROI type).
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
            bin_width(int, optional): Width of the bins for the histograms. If not provided, will use the 
                default number of bins in the method 
                :ref:`pandas.DataFrame.hist <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.hist.html>`: 10 bins.
            hist_range(list, optional): Range of the histograms. If empty, will use the minimum and maximum values.
            nifti (bool, optional): Set to True if the scans are nifti files. Defaults to False.
            save (bool, optional): If True, will save the results in a json file. Defaults to False.

        Returns:
            None
        """
        # Initialization
        path_study = Path.cwd()

        # Load params
        if not self.paths._path_pre_checks_settings:
            if not wildcards_dimensions or not wildcards_window:
                raise ValueError("path to pre-checks settings is None.\
                    wildcards_dimensions and wildcards_window need to be specified")
        else:
            settings = self.paths._path_pre_checks_settings
            settings = load_json(settings)
            settings = settings['pre_radiomics_checks']  

            # Setting up paths
            if 'path_save_checks' in settings and settings['path_save_checks']:
                self.paths._path_save_checks = Path(settings['path_save_checks']) 
            if 'path_csv' in settings and settings['path_csv']:
                self.paths._path_csv = Path(settings['path_csv']) 

            # Wildcards of groups of files to analyze for dimensions in path_data.
            # See for example: https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html
            # Keep the cell empty if no dimension checks are to be performed.
            if not wildcards_dimensions:
                wildcards_dimensions = []
                for i in range(len(settings['wildcards_dimensions'])):
                    wildcards_dimensions.append(settings['wildcards_dimensions'][i])

            # ROI intensity window checks params
            if not wildcards_window:
                wildcards_window = []
                for i in range(len(settings['wildcards_window'])):
                    wildcards_window.append(settings['wildcards_window'][i])

        # PRE-RADIOMICS CHECKS
        if not self.paths._path_save_checks:
            if (path_study / 'checks').exists():
                self.paths._path_save_checks = Path(path_study / 'checks')
            else:
                os.mkdir(path_study / 'checks')
                self.paths._path_save_checks = Path(path_study / 'checks')
        else:
            if self.paths._path_save_checks.name != 'checks':
                if (self.paths._path_save_checks / 'checks').exists():
                    self.paths._path_save_checks /= 'checks'
                else:
                    os.mkdir(self.paths._path_save_checks / 'checks')
                    self.paths._path_save_checks = Path(self.paths._path_save_checks / 'checks')

        # Initializing plotting params
        plt.rcParams["figure.figsize"] = (20,20)
        plt.rcParams.update({'font.size': 22})
        
        start = time()
        print('\n\n************************* PRE-RADIOMICS CHECKS *************************', end='')

        # 1. PRE-RADIOMICS CHECKS -- DIMENSIONS
        start1 = time()
        print('\n--> PRE-RADIOMICS CHECKS -- DIMENSIONS ... ', end='')
        self.__pre_radiomics_checks_dimensions(
                                        path_data, 
                                        wildcards_dimensions, 
                                        min_percentile, 
                                        max_percentile,
                                        save)
        print('DONE', end='')
        time1 = f"{time() - start1:.2f}"
        print(f'\nElapsed time: {time1} sec', end='')

        # 2. PRE-RADIOMICS CHECKS - WINDOW
        start2 = time()
        print('\n\n--> PRE-RADIOMICS CHECKS -- WINDOW ... \n', end='')
        self.__pre_radiomics_checks_window(
                                        path_data, 
                                        wildcards_window, 
                                        path_csv,
                                        min_percentile, 
                                        max_percentile,
                                        bin_width,
                                        hist_range,
                                        nifti,
                                        save)
        print('DONE', end='')
        time2 = f"{time() - start2:.2f}"
        print(f'\nElapsed time: {time2} sec', end='')

        time_elapsed = f"{time() - start:.2f}"
        print(f'\n\n--> TOTAL TIME FOR PRE-RADIOMICS CHECKS: {time_elapsed} seconds')
        print('-------------------------------------------------------------------------------------')

    def perform_mr_imaging_summary(self, 
                                wildcards_scans: List[str],
                                path_data: Path = None,
                                path_save_checks: Path = None,
                                min_percentile: float = 0.05,
                                max_percentile: float = 0.95
                                ) -> None:
        """
        Summarizes MRI imaging acquisition parameters. Plots summary histograms
        for different dimensions and saves all acquisition parameters locally in JSON files.

        Args:
            wildcards_scans (List[str]): List of wildcards that determines the scans 
                that will be analyzed (Only MRI scans will be analyzed). You can learn more about wildcards in
                `this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`_.
                For example: ``[\"STS*.MRscan.npy\"]``.
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            path_save_checks (Path, optional): Path where to save the checks, if not specified will use the one 
                in the current instance.
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
        
        Returns:
            None.
        """
        # Initializing data structures
        class param:
            dates = []
            manufacturer = []
            scanning_sequence = []
            class years:
                data = []

            class fieldStrength:
                data = []

            class repetitionTime:
                data = []

            class echoTime:
                data = []

            class inversionTime:
                data = []

            class echoTrainLength:
                data = []

            class flipAngle:
                data = []

            class numberAverages:
                data = []

            class xyDim:
                data = []

            class zDim:
                data = []
        
        if len(wildcards_scans) == 0:
            print('wildcards_scans is empty')

        # wildcards checks:
        no_mr_scan = True
        for wildcard in wildcards_scans:
            if 'MRscan' in wildcard:
                no_mr_scan = False
        if no_mr_scan:
            raise ValueError(f"wildcards: {wildcards_scans} does not include MR scans. (Only MR scans are supported)")
        
        # Initialization
        if path_data is None:
            if self.paths._path_save:
                path_data = Path(self.paths._path_save)
            else:
                print("No path to data was given and path save is None.")
                return 0
        
        if not path_save_checks:
            if self.paths._path_save_checks:
                path_save_checks = Path(self.paths._path_save_checks)
            else:
                if (Path(os.getcwd()) / "checks").exists():
                    path_save_checks = Path(os.getcwd()) / "checks"
                else:
                    path_save_checks = (Path(os.getcwd()) / "checks").mkdir()
        # Looping through all the different wildcards
        for i in tqdm(range(len(wildcards_scans))):
            wildcard = wildcards_scans[i]
            file_paths = get_file_paths(path_data, wildcard)
            n_files = len(file_paths)
            param.dates = np.zeros(n_files)
            param.years.data = np.zeros((n_files, 1))
            param.years.data = np.multiply(param.years.data, np.NaN)
            param.manufacturer = [None] * n_files
            param.scanning_sequence = [None] * n_files
            param.fieldStrength.data = np.zeros((n_files, 1))
            param.fieldStrength.data = np.multiply(param.fieldStrength.data, np.NaN)
            param.repetitionTime.data = np.zeros((n_files, 1))
            param.repetitionTime.data = np.multiply(param.repetitionTime.data, np.NaN)
            param.echoTime.data = np.zeros((n_files, 1))
            param.echoTime.data = np.multiply(param.echoTime.data, np.NaN)
            param.inversionTime.data = np.zeros((n_files, 1))
            param.inversionTime.data = np.multiply(param.inversionTime.data, np.NaN)
            param.echoTrainLength.data = np.zeros((n_files, 1))
            param.echoTrainLength.data = np.multiply(param.echoTrainLength.data, np.NaN)
            param.flipAngle.data = np.zeros((n_files, 1))
            param.flipAngle.data = np.multiply(param.flipAngle.data, np.NaN)
            param.numberAverages.data = np.zeros((n_files, 1))
            param.numberAverages.data = np.multiply(param.numberAverages.data, np.NaN)
            param.xyDim.data = np.zeros((n_files, 1))
            param.xyDim.data = np.multiply(param.xyDim.data, np.NaN)
            param.zDim.data = np.zeros((n_files, 1))
            param.zDim.data = np.multiply(param.zDim.data, np.NaN)
            
            # Loading and recording data
            for f in tqdm(range(n_files)):
                file = file_paths[f]

                #Open file for warning
                try:
                    warn_file = open(path_save_checks / 'imaging_summary_mr_warnings.txt', 'a')
                except IOError:
                    print("Could not open warning file")

                # Loading Data
                try:
                    print(f'\nCurrently working on: {file}', file = warn_file)
                    with open(path_data / file, 'rb') as fe: medscan = pickle.load(fe)

                    # Example of DICOM header
                    info = medscan.dicomH[1]
                    # Recording dates (info.AcquistionDates)
                    try:
                        param.dates[f] = info.AcquisitionDate
                    except AttributeError:
                        param.dates[f] = info.StudyDate
                    # Recording years
                    try:
                        y = str(param.dates[f])  # Only the first four characters represent the years
                        param.years.data[f] = y[0:4]
                    except Exception as e:
                        print(f'Cannot read years of: {file}. Error: {e}', file = warn_file)
                    # Recording manufacturers
                    try:
                        param.manufacturer[f] = info.Manufacturer
                    except Exception as e:
                        print(f'Cannot read manufacturer of: {file}. Error: {e}', file = warn_file)
                    # Recording scanning sequence
                    try:
                        param.scanning_sequence[f] = info.scanning_sequence
                    except Exception as e:
                        print(f'Cannot read scanning sequence of: {file}. Error: {e}', file = warn_file)
                    # Recording field strength
                    try:
                        param.fieldStrength.data[f] = info.MagneticFieldStrength
                    except Exception as e:
                        print(f'Cannot read field strength of: {file}. Error: {e}', file = warn_file)
                    # Recording repetition time
                    try:
                        param.repetitionTime.data[f] = info.RepetitionTime
                    except Exception as e:
                        print(f'Cannot read repetition time of: {file}. Error: {e}', file = warn_file)
                    # Recording echo time
                    try:
                        param.echoTime.data[f] = info.EchoTime
                    except Exception as e:
                        print(f'Cannot read echo time of: {file}. Error: {e}', file = warn_file)
                    # Recording inversion time
                    try:
                        param.inversionTime.data[f] = info.InversionTime
                    except Exception as e:
                        print(f'Cannot read inversion time of: {file}. Error: {e}', file = warn_file)
                    # Recording echo train length
                    try:
                        param.echoTrainLength.data[f] = info.EchoTrainLength
                    except Exception as e:
                        print(f'Cannot read echo train length of: {file}. Error: {e}', file = warn_file)
                    # Recording flip angle
                    try:
                        param.flipAngle.data[f] = info.FlipAngle
                    except Exception as e:
                        print(f'Cannot read flip angle of: {file}. Error: {e}', file = warn_file)
                    # Recording number of averages
                    try:
                        param.numberAverages.data[f] = info.NumberOfAverages
                    except Exception as e:
                        print(f'Cannot read number averages of: {file}. Error: {e}', file = warn_file)
                    # Recording xy spacing
                    try:
                        param.xyDim.data[f] = medscan.data.volume.spatialRef.PixelExtentInWorldX
                    except Exception as e:
                        print(f'Cannot read x spacing of: {file}. Error: {e}', file = warn_file)
                    # Recording z spacing
                    try:
                        param.zDim.data[f] = medscan.data.volume.spatialRef.PixelExtentInWorldZ
                    except Exception as e:
                        print(f'Cannot read z spacing of: {file}', file = warn_file)
                except Exception as e:
                    print(f'Cannot read file: {file}. Error: {e}', file = warn_file)

                warn_file.close()

            # Summarize data
                # Summarizing years
            df_years = pd.DataFrame(param.years.data, 
                                    columns=['years']).describe(percentiles=[min_percentile, max_percentile], 
                                    include='all')
                # Summarizing field strength
            df_fs = pd.DataFrame(param.fieldStrength.data, 
                                columns=['fieldStrength']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing  repetition time
            df_rt = pd.DataFrame(param.repetitionTime.data, 
                                columns=['repetitionTime']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing echo time
            df_et = pd.DataFrame(param.echoTime.data, 
                                columns=['echoTime']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing inversion time
            df_it = pd.DataFrame(param.inversionTime.data, 
                                columns=['inversionTime']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing echo train length
            df_etl = pd.DataFrame(param.echoTrainLength.data, 
                                columns=['echoTrainLength']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing flip  angle
            df_fa = pd.DataFrame(param.flipAngle.data, 
                                columns=['flipAngle']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing number of  averages
            df_na = pd.DataFrame(param.numberAverages.data, 
                                columns=['numberAverages']).describe(percentiles=[min_percentile, max_percentile], 
                                include='all')
                # Summarizing xy-spacing
            df_xy = pd.DataFrame(param.xyDim.data, 
                                columns=['xyDim'])
                # Summarizing z-spacing
            df_z = pd.DataFrame(param.zDim.data, 
                                columns=['zDim'])

            # Plotting xy-spacing histogram
            ax = df_xy.hist(column='xyDim')
            min_quant, max_quant, average = df_xy.quantile(min_percentile), df_xy.quantile(max_percentile), param.xyDim.data.mean()
            for x in ax[0]:
                x.axvline(min_quant.xyDim, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.xyDim, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(average, linestyle='solid', color='gold', label=f"Average: {float(average):.3f}")
                x.grid(False)
                plt.title(f"MR xy-spacing imaging summary for {wildcard}")
                plt.legend()
                plt.show()
            
            # Plotting z-spacing histogram
            ax = df_z.hist(column='zDim')
            min_quant, max_quant, average = df_z.quantile(min_percentile), df_z.quantile(max_percentile), param.zDim.data.mean()
            for x in ax[0]:
                x.axvline(min_quant.zDim, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.zDim, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(average, linestyle='solid', color='gold', label=f"Average: {float(average):.3f}")
                x.grid(False)
                plt.title(f"MR z-spacing imaging summary for {wildcard}")
                plt.legend()
                plt.show()
            
            # Summarizing xy-spacing
            df_xy = df_xy.describe(percentiles=[min_percentile, max_percentile], include='all')
            # Summarizing  z-spacing
            df_z = df_z.describe(percentiles=[min_percentile, max_percentile], include='all')
            
            # Saving data
            name_save = wildcard.replace('*', '').replace('.npy', '')
            save_name = 'imagingSummary__' + name_save + ".json"
            df_all = [df_years, df_fs, df_rt, df_et, df_it, df_etl, df_fa, df_na, df_xy, df_z]
            df_all = df_all[0].join(df_all[1:])
            df_all.to_json(path_save_checks / save_name, orient='columns', indent=4)

    def perform_ct_imaging_summary(self, 
                                wildcards_scans: List[str],
                                path_data: Path = None,
                                path_save_checks: Path = None,
                                min_percentile: float = 0.05,
                                max_percentile: float = 0.95
                                ) -> None:
        """
        Summarizes CT imaging acquisition parameters. Plots summary histograms
        for different dimensions and saves all acquisition parameters locally in JSON files.

        Args:
            wildcards_scans (List[str]): List of wildcards that determines the scans 
                that will be analyzed (Only MRI scans will be analyzed). You can learn more about wildcards in
                `this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`_.
                For example: ``[\"STS*.CTscan.npy\"]``.
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            path_save_checks (Path, optional): Path where to save the checks, if not specified will use the one 
                in the current instance.
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
        
        Returns:
            None.
        """

        class param:
            manufacturer = []
            dates = []
            kernel = []

            class years:
                data = []
            class voltage:
                data = []
            class exposure:
                data = []
            class xyDim:
                data = []
            class zDim:
                data = []

        if len(wildcards_scans) == 0:
            print('wildcards_scans is empty')

        # wildcards checks:
        no_mr_scan = True
        for wildcard in wildcards_scans:
            if 'CTscan' in wildcard:
                no_mr_scan = False
        if no_mr_scan:
            raise ValueError(f"wildcards: {wildcards_scans} does not include CT scans. (Only CT scans are supported)")

        # Initialization
        if path_data is None:
            if self.paths._path_save:
                path_data = Path(self.paths._path_save)
            else:
                print("No path to data was given and path save is None.")
                return 0
        
        if not path_save_checks:
            if self.paths._path_save_checks:
                path_save_checks = Path(self.paths._path_save_checks)
            else:
                if (Path(os.getcwd()) / "checks").exists():
                    path_save_checks = Path(os.getcwd()) / "checks"
                else:
                    path_save_checks = (Path(os.getcwd()) / "checks").mkdir()

        # Looping through all the different wildcards
        for i in tqdm(range(len(wildcards_scans))):
            wildcard = wildcards_scans[i]
            file_paths = get_file_paths(path_data, wildcard)
            n_files = len(file_paths)
            param.dates = np.zeros(n_files)
            param.years.data = np.zeros(n_files)
            param.years.data = np.multiply(param.years.data, np.NaN)
            param.manufacturer = [None] * n_files
            param.voltage.data = np.zeros(n_files)
            param.voltage.data = np.multiply(param.voltage.data, np.NaN)
            param.exposure.data = np.zeros(n_files)
            param.exposure.data = np.multiply(param.exposure.data, np.NaN)
            param.kernel = [None] * n_files
            param.xyDim.data = np.zeros(n_files)
            param.xyDim.data = np.multiply(param.xyDim.data, np.NaN)
            param.zDim.data = np.zeros(n_files)
            param.zDim.data = np.multiply(param.zDim.data, np.NaN)
        
            # Loading and recording data
            for f in tqdm(range(n_files)):
                file = file_paths[f]

                # Open file for warning
                try:
                    warn_file = open(path_save_checks / 'imaging_summary_ct_warnings.txt', 'a')
                except IOError:
                    print("Could not open file")

                # Loading Data
                try:
                    with open(path_data / file, 'rb') as fe: medscan = pickle.load(fe)
                    print(f'Currently working on: {file}', file=warn_file)
                    
                    # DICOM header
                    info = medscan.dicomH[1]

                    # Recording dates
                    try:
                        param.dates[f] = info.AcquisitionDate
                    except AttributeError:
                        param.dates[f] = info.StudyDate
                        # Recording years
                    try:
                        years = str(param.dates[f])  # Only the first four characters represent the years
                        param.years.data[f] = years[0:4]
                    except Exception as e:
                        print(f'Cannot read dates of : {file}. Error: {e}', file=warn_file)
                        # Recording manufacturers
                    try:
                        param.manufacturer[f] = info.Manufacturer
                    except Exception as e:
                        print(f'Cannot read Manufacturer of: {file}. Error: {e}', file=warn_file)
                        # Recording voltage
                    try:
                        param.voltage.data[f] = info.KVP
                    except Exception as e:
                        print(f'Cannot read voltage of: {file}. Error: {e}', file=warn_file)
                        # Recording exposure
                    try:
                        param.exposure.data[f] = info.Exposure
                    except Exception as e:
                        print(f'Cannot read exposure of: {file}. Error: {e}', file=warn_file)
                        # Recording reconstruction kernel
                    try:
                        param.kernel[f] = info.ConvolutionKernel
                    except Exception as e:
                        print(f'Cannot read Kernel of: {file}. Error: {e}', file=warn_file)
                        # Recording xy spacing
                    try:
                        param.xyDim.data[f] = medscan.data.volume.spatialRef.PixelExtentInWorldX
                    except Exception as e:
                        print(f'Cannot read x spacing of: {file}. Error: {e}', file=warn_file)
                        # Recording z spacing
                    try:
                        param.zDim.data[f] = medscan.data.volume.spatialRef.PixelExtentInWorldZ
                    except Exception as e:
                        print(f'Cannot read z spacing of: {file}. Error: {e}', file=warn_file)
                except Exception as e:
                    print(f'Cannot load file: {file}', file=warn_file)

                warn_file.close()

            # Summarize data
                # Summarizing years
            df_years = pd.DataFrame(param.years.data, columns=['years']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarizing voltage
            df_voltage = pd.DataFrame(param.voltage.data, columns=['voltage']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarizing exposure
            df_exposure = pd.DataFrame(param.exposure.data, columns=['exposure']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarizing kernel
            df_kernel = pd.DataFrame(param.kernel, columns=['kernel']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarize xy spacing
            df_xy = pd.DataFrame(param.xyDim.data, columns=['xyDim']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarize z spacing
            df_z = pd.DataFrame(param.zDim.data, columns=['zDim']).describe(percentiles=[min_percentile, max_percentile], include='all')
                # Summarizing xy-spacing
            df_xy = pd.DataFrame(param.xyDim.data, columns=['xyDim'])
                # Summarizing z-spacing
            df_z = pd.DataFrame(param.zDim.data, columns=['zDim'])

            # Plotting xy-spacing histogram
            ax = df_xy.hist(column='xyDim')
            min_quant, max_quant, average = df_xy.quantile(min_percentile), df_xy.quantile(max_percentile), param.xyDim.data.mean()
            for x in ax[0]:
                x.axvline(min_quant.xyDim, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.xyDim, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(average, linestyle='solid', color='gold', label=f"Average: {float(average):.3f}")
                x.grid(False)
                plt.title(f"CT xy-spacing imaging summary for {wildcard}")
                plt.legend()
                plt.show()
            
            # Plotting z-spacing histogram
            ax = df_z.hist(column='zDim')
            min_quant, max_quant, average = df_z.quantile(min_percentile), df_z.quantile(max_percentile), param.zDim.data.mean()
            for x in ax[0]:
                x.axvline(min_quant.zDim, linestyle=':', color='r', label=f"Min Percentile: {float(min_quant):.3f}")
                x.axvline(max_quant.zDim, linestyle=':', color='g', label=f"Max Percentile: {float(max_quant):.3f}")
                x.axvline(average, linestyle='solid', color='gold', label=f"Average: {float(average):.3f}")
                x.grid(False)
                plt.title(f"CT z-spacing imaging summary for {wildcard}")
                plt.legend()
                plt.show()
            
            # Summarizing xy-spacing
            df_xy = df_xy.describe(percentiles=[min_percentile, max_percentile], include='all')
            # Summarizing  z-spacing
            df_z = df_z.describe(percentiles=[min_percentile, max_percentile], include='all')

            # Saving data
            name_save = wildcard.replace('*', '').replace('.npy', '')
            save_name = 'imagingSummary__' + name_save + ".json"
            df_all = [df_years, df_voltage, df_exposure, df_kernel, df_xy, df_z]
            df_all = df_all[0].join(df_all[1:])
            df_all.to_json(path_save_checks / save_name, orient='columns', indent=4)

    def perform_imaging_summary(self, 
                                wildcards_scans: List[str],
                                path_data: Path = None,
                                path_save_checks: Path = None,
                                min_percentile: float = 0.05,
                                max_percentile: float = 0.95
                                ) -> None:
        """
        Summarizes CT and MR imaging acquisition parameters. Plots summary histograms
        for different dimensions and saves all acquisition parameters locally in JSON files.

        Args:
            wildcards_scans (List[str]): List of wildcards that determines the scans 
                that will be analyzed (CT and MRI scans will be analyzed). You can learn more about wildcards in
                `this link <https://www.linuxtechtips.com/2013/11/how-wildcards-work-in-linux-and-unix.html>`_.
                For example: ``[\"STS*.CTscan.npy\", \"STS*.MRscan.npy\"]``.
            path_data (Path, optional): Path to the MEDscan objects, if not specified will use ``path_save`` from the 
                inner-class ``Paths`` in the current instance.
            path_save_checks (Path, optional): Path where to save the checks, if not specified will use the one 
                in the current instance.
            min_percentile (float, optional): Minimum percentile to use for the histograms. Defaults to 0.05.
            max_percentile (float, optional): Maximum percentile to use for the histograms. Defaults to 0.95.
        
        Returns:
            None.
        """
        # MR imaging summary
        wildcards_scans_mr = [wildcard for wildcard in wildcards_scans if 'MRscan' in wildcard]
        if len(wildcards_scans_mr) == 0:
            print("Cannot perform imaging summary for MR, no MR scan wildcard was given! ")
        else:
            self.perform_mr_imaging_summary(
                                wildcards_scans_mr, 
                                path_data, 
                                path_save_checks, 
                                min_percentile, 
                                max_percentile)
        # CT imaging summary
        wildcards_scans_ct = [wildcard for wildcard in wildcards_scans if 'CTscan' in wildcard]
        if len(wildcards_scans_ct) == 0:
            print("Cannot perform imaging summary for CT, no CT scan wildcard was given! ")
        else:
            self.perform_ct_imaging_summary(
                                wildcards_scans_ct, 
                                path_data, 
                                path_save_checks, 
                                min_percentile, 
                                max_percentile)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Dict
import numpy as np


def get_radiomic_names(roi_names: np.array,
                       roi_type: str) -> Dict:
    """Generates radiomics names using ``roi_names`` and ``roi_types``.

    Args:
        roi_names (np.array): array of the ROI names.
        roi_type(str): string of the ROI.

    Returns:
        dict: dict with the radiomic names
    """

    n_names = np.size(roi_names)[0]
    radiomic_names = [0] * n_names
    for n in range(0, n_names):
        radiomic_names[n] = roi_names[n, 0]+'__'+roi_names[n, 1] + \
            '('+roi_type+').'+roi_names[n, 2]+'.npy'

    return radiomic_names



import platform
import re
from itertools import combinations, product
from pathlib import Path
from typing import Dict, List

import pandas as pd

from ..utils.get_institutions_from_ids import get_institutions_from_ids
from ..utils.json_utils import load_json, posix_to_string, save_json
from .ml_utils import cross_validation_split, get_stratified_splits


class DesignExperiment:
    def __init__(self, path_study: Path, path_settings: Path, experiment_label: str) -> None:
        """
        Constructor of the class DesignExperiment.

        Args:
            path_study (Path): Path to the main study folder where the outcomes, 
                learning patients and holdout patients dictionaries are found.
            path_settings (Path): Path to the settings folder.
            experiment_label (str): String specifying the label to attach to a given learning experiment in 
                "path_experiments". This label will be attached to the ml__$experiments_label$.json file as well
                as the learn__$experiment_label$ folder. This label is used to keep track of different experiments 
                with different settings (e.g. radiomics, scans, machine learning algorithms, etc.).
        
        Returns:
            None
        """
        self.path_study = Path(path_study)
        self.path_settings = Path(path_settings)
        self.experiment_label = str(experiment_label)
        self.path_ml_object = None

    def __create_folder_and_content(
            self, 
            path_learn: Path,
            run_name: str, 
            patients_train: List,
            patients_test: List, 
            ml_path: Path
        ) -> List:
        """
        Creates json files needed for a given run
        
        Args:
            path_learn (Path): path to the main learning folder containing information about the training and test set.
            run_name (str): name for a given run.
            patients_train (List): list of patients in the training set.
            patients_test (List): list of patients in the test set.
            ml_path (Path): path to the given run.
        
        Returns:
            List: list of paths to the given run.
        """
        paths_ml = dict()
        path_run = path_learn / run_name
        Path.mkdir(path_run, exist_ok=True)
        path_train = path_run / 'patientsTrain.json'
        path_test = path_run / 'patientsTest.json'
        save_json(path_train, sorted(patients_train))
        save_json(path_test, sorted(patients_test))
        paths_ml['patientsTrain'] = path_train
        paths_ml['patientsTest'] = path_test
        paths_ml['outcomes'] = self.path_study / 'outcomes.csv'
        paths_ml['ml'] = self.path_ml_object
        paths_ml['results'] = path_run / 'run_results.json'
        path_file = path_run / 'paths_ml.json'
        paths_ml = posix_to_string(paths_ml)
        ml_path.append(path_file)
        save_json(path_file, paths_ml)

        return ml_path

    def generate_learner_dict(self) -> dict:
        """
        Generates a dictionary containing all the settings for the learning experiment.
        
        Returns:
            dict: Dictionary containing all the settings for the learning experiment.
        """
        ml_options = dict()

        # operating system
        ml_options['os'] = platform.system()

        # design experiment settings
        ml_options['design'] = self.path_settings / 'ml_design.json'
        # check if file exist:
        if not ml_options['design'].exists():
            raise FileNotFoundError(f"File {ml_options['design']} does not exist.")

        # ML run settings
        run = dict()
        ml_options['run'] = run

        # Machine learning settings
        ml_options['settings'] = self.path_settings / 'ml_settings.json'
        # check if file exist:
        if not ml_options['settings'].exists():
            raise FileNotFoundError(f"File {ml_options['settings']} does not exist.")

        # variables settings
        ml_options['variables'] = self.path_settings / 'ml_variables.json'
        # check if file exist:
        if not ml_options['variables'].exists():
            raise FileNotFoundError(f"File {ml_options['variables']} does not exist.")

        # ML algorithms settings
        ml_options['algorithms'] = self.path_settings / 'ml_algorithms.json'
        # check if file exist:
        if not ml_options['algorithms'].exists():
            raise FileNotFoundError(f"File {ml_options['algorithms']} does not exist.")

        # Data cleaning settings
        ml_options['datacleaning'] = self.path_settings / 'ml_datacleaning.json'
        # check if file exist:
        if not ml_options['datacleaning'].exists():
            raise FileNotFoundError(f"File {ml_options['datacleaning']} does not exist.")

        # Normalization settings
        ml_options['normalization'] = self.path_settings / 'ml_normalization.json'
        # check if file exist:
        if not ml_options['normalization'].exists():
            raise FileNotFoundError(f"File {ml_options['normalization']} does not exist.")

        # Feature set reduction settings
        ml_options['fSetReduction'] = self.path_settings / 'ml_fset_reduction.json'
        # check if file exist:
        if not ml_options['fSetReduction'].exists():
            raise FileNotFoundError(f"File {ml_options['fSetReduction']} does not exist.")

        # Experiment label check
        if self.experiment_label == "":
            raise ValueError("Experiment label is empty. Class was not initialized properly.")
        
        # save all the ml options and return the path to the saved file
        name_save_options = 'ml_options_' + self.experiment_label + '.json'
        path_ml_options = self.path_settings / name_save_options
        ml_options = posix_to_string(ml_options)
        save_json(path_ml_options, ml_options)
        
        return path_ml_options

    def fill_learner_dict(self, path_ml_options: Path) -> Path:
        """
        Fills the main expirement dictionary from the settings in the different json files. 
        This main dictionary will hold all the settings for the data processing and learning experiment.

        Args:
            path_ml_options (Path): Path to the ml_options json file for the experiment.
        
        Returns:
            Path: Path to the learner object.
        """
        # Initialization
        all_datacleaning = list()
        all_normalization = list()
        all_fset_reduction = list()

        # Load ml options dict
        ml_options = load_json(path_ml_options)
        options = ml_options.keys()

        # Design options
        ml = dict()
        ml['design'] = load_json(ml_options['design'])

        # ML run options
        ml['run'] = ml_options['run']

        # Machine learning options
        if 'settings' in options:
            ml['settings'] = load_json(ml_options['settings'])

        # Machine learning variables
        if 'variables' in options:
            ml['variables'] = dict()
            var_options = load_json(ml_options['variables'])
            fields = list(var_options.keys())
            vars = [(idx, s) for idx, s in enumerate(fields) if re.match(r"^var[0-9]{1,}$", s)]
            var_names = [var[1] for var in vars]  # list of var names
            
            # For each variable, organize the option in the ML dictionary
            for (idx, var) in vars:
                vars_dict = dict()
                vars_dict[var] = var_options[var]
                var_struct = var_options[var]
                
                # Radiomics variables
                if 'radiomics' in var_struct['nameType'].lower():
                    # Get radiomics features in workspace
                    if 'settofeatures' in var_struct['path'].lower():
                        name_folder = re.match(r"setTo(.*)inWorkspace", var_struct['path']).group(1)
                        path_features = self.path_study / name_folder
                    # Get radiomics features in path provided in the dictionary by the user 
                    else:
                        path_features = var_struct['path']
                    scans = var_struct['scans'] # list of imaging sequences
                    rois = var_struct['rois'] # list of roi labels
                    im_spaces = var_struct['imSpaces'] # list of image spaces (filterd and original)
                    use_combinations = var_struct['use_combinations'] # boolean to use combinations of scans and im_spaces
                    if use_combinations:
                        all_combinations = []
                        scans = list(var_struct['combinations'].keys())
                        for scan in scans:
                            im_spaces = list(var_struct['combinations'][scan])
                            all_combinations += list(product([scan], rois, im_spaces))
                    else:
                        all_combinations = list(product(scans, rois, im_spaces))
                    
                    # Initialize dict to hold all paths to radiomics features (csv and txt files)
                    path = dict() 
                    for idx, (scan, roi, im_space) in enumerate(all_combinations):
                        rad_tab_x = {}
                        name_tab = 'radTab' + str(idx+1)
                        radiomics_table_name = 'radiomics__' + scan + '(' + roi + ')__' + im_space
                        rad_tab_x['csv'] = path_features / (radiomics_table_name + '.csv')
                        rad_tab_x['txt'] = path_features / (radiomics_table_name + '.txt')
                        rad_tab_x['type'] = path_features / (scan + '(' + roi + ')__' + im_space)
                        
                        # check if file exist
                        if not rad_tab_x['csv'].exists():
                            raise FileNotFoundError(f"File {rad_tab_x['csv']} does not exist.")
                        if not rad_tab_x['txt'].exists():
                            raise FileNotFoundError(f"File {rad_tab_x['txt']} does not exist.")
                        
                        path[name_tab] = rad_tab_x
                    
                    # Add path to ml dict for the current variable
                    vars_dict[var]['path'] = path
                    
                    # Add to ml dict for the current variable
                    ml['variables'].update(vars_dict)
                
                # Clinical or other variables (For ex: Volume)
                else:
                    # get path to csv file of features
                    if not var_struct['path']:
                        if var_options['pathCSV'] == 'setToCSVinWorkspace':
                            path_csv = self.path_study / 'CSV'
                        else:
                            path_csv = var_options['pathCSV']
                        var_struct['path'] = path_csv / var_struct['nameFile']
                
                # Add to ml dict for the current variable
                ml['variables'].update(vars_dict)

                # Initialize data processing methods
                if 'var_datacleaning' in var_struct.keys():
                    all_datacleaning.append(var_struct['var_datacleaning'])
                if 'var_normalization' in var_struct.keys():
                    all_normalization.append((var_struct['var_normalization']))
                if 'var_fSetReduction' in var_struct.keys():
                    all_fset_reduction.append(var_struct['var_fSetReduction']['method'])

            # Combinations of variables
            if 'combinations' in var_options.keys():
                if var_options['combinations'] == ['all']:  # Combine all variables
                    combs = [comb for i in range(len(vars)) for comb in combinations(var_names, i+1)]
                    combstrings = ['_'.join(elt) for elt in combs]
                    ml['variables']['combinations'] = combstrings
                else:
                    ml['variables']['combinations'] = var_options['combinations']
            
            # Varibles to use for ML
            ml['variables']['varStudy'] = var_options['varStudy']

        # ML algorithms
        if 'algorithms' in options:
            algorithm = ml['settings']['algorithm']
            algorithms = load_json(ml_options['algorithms'])
            ml['algorithms'] = {}
            ml['algorithms'][algorithm] = algorithms[algorithm]

        # ML data processing methods and its options
        for (method, method_list) in [
                ('datacleaning', all_datacleaning),
                ('normalization', all_normalization), 
                ('fSetReduction', all_fset_reduction)
            ]:
            # Skip if no method is selected
            if all(v == "" for v in method_list):
                continue
            if method in options:
                # Add algorithm specific methods
                if method in ml['settings'].keys():
                    method_list.append(ml['settings'][method])
                    method_list = list(set(method_list))  # to only get unique values of all_datacleaning
                method_options = load_json(ml_options[method]) # load json file of each method
                if method == 'normalization' and 'combat' in method_list:
                    ml[method] = 'combat'
                    continue
                ml[method] = dict()
                for name in list(set(method_list)):
                    if name != "":
                        ml[method][name] = method_options[name]

        # Save the ML dictionary
        if self.experiment_label == "":
            raise ValueError("Experiment label is empty. Class was not initialized properly.")
        path_ml_object = self.path_study / f'ml__{self.experiment_label}.json'
        ml = posix_to_string(ml)    # Convert all paths to string
        save_json(path_ml_object, ml)
        
        # return ml
        return path_ml_object

    def create_experiment(self, ml: dict = None) -> Dict:
        """
        Create the machine learning experiment dictionary, organizes each test/split information in a seperate folder.

        Args:
            ml (dict, optional): Dictionary containing all the machine learning settings. Defaults to None.
        
        Returns:
            Dict: Dictionary containing all the organized machine learning settings.
        """
        # Initialization
        ml_path = list()
        ml = load_json(self.path_ml_object) if ml is None else ml

        # Learning set
        patients_learn = load_json(self.path_study / 'patientsLearn.json')
        
        # Outcomes table
        outcomes_table = pd.read_csv(self.path_study / 'outcomes.csv', index_col=0)

        # keep only patients in learn set and outcomes table
        patients_to_keep = list(filter(lambda x: x in patients_learn, outcomes_table.index.values.tolist()))
        outcomes_table = outcomes_table.loc[patients_to_keep]

        # Get the "experiment label" from ml__$experiment_label$.json
        if self.experiment_label:
            experiment_label = self.experiment_label
        else:
            experiment_label = Path(self.path_ml_object).name[4:-5]

        # Create the folder for the training and testing sets (machine learning) information
        name_learn = 'learn__' + experiment_label
        path_learn = Path(self.path_study) / name_learn
        Path.mkdir(path_learn, exist_ok=True)

        # Getting the type of test_sets
        test_sets_types = ml['design']['testSets']

        # Creating the sets for the different machine learning runs
        for type_set in test_sets_types:
            # Random splits
            if type_set.lower() == 'random':
                # Get the experiment options for the sets
                random_info = ml['design'][type_set]
                method = random_info['method']
                n_splits = random_info['nSplits']
                stratify_institutions = random_info['stratifyInstitutions']
                test_proportion = random_info['testProportion']
                seed = random_info['seed']
                if method == 'SubSampling':
                    # Get the training and testing sets
                    patients_train, patients_test = get_stratified_splits(
                        outcomes_table, n_splits, 
                        test_proportion, seed, 
                        stratify_institutions
                    )

                    # If patients are not in a list
                    if type(patients_train) != list and not hasattr((patients_train), "__len__"):
                        patients_train = [patients_train]
                        patients_test = [patients_test]
                    
                    for i in range(n_splits):
                        # Create a folder for each split/run
                        run_name = "test__{0:03}".format(i+1)
                        ml_path = self.__create_folder_and_content(
                            path_learn, 
                            run_name, 
                            patients_train[i], 
                            patients_test[i], 
                            ml_path
                        )
            # Institutions-based splits
            elif type_set.lower() == 'institutions':
                # Get institutions run info
                patient_ids = pd.Series(outcomes_table.index)
                institution_cat_vector = get_institutions_from_ids(patient_ids)
                institution_cats = list(set(institution_cat_vector))
                n_institution = len(institution_cats)
                # The 'Institutions' argument only make sense if n_institutions > 1
                if n_institution > 1:
                    for i in range(n_institution):
                        cat = institution_cats[i]
                        patients_train = [elt for elt in patient_ids if cat not in elt]
                        patients_test = [elt for elt in patient_ids if cat in elt]
                        run_name = f"test__{cat}"
                        # Create a folder for each split/run
                        ml_path = self.__create_folder_and_content(
                            path_learn,
                            run_name,
                            patients_train,
                            patients_test,
                            ml_path
                        )
                    if n_institution > 2:
                        size_inst = list()
                        for i in range(n_institution):
                            cat = institution_cats[i]
                            size_inst.append(sum([1 if cat in elt else 0 for elt in institution_cat_vector]))
                        ind_max = size_inst.index(max(size_inst))
                        str_test = list()
                        for i in range(n_institution):
                            if i != ind_max:
                                cat = institution_cats[i]
                                str_test.append(cat)
                        cat = institution_cats[ind_max]
                        patients_train = [elt for elt in patient_ids if cat in elt]
                        patients_test = [elt for elt in patient_ids if cat not in elt]
                        run_name = f"test__{'_'.join(str_test)}"
                        # Create a folder for each split/run
                        ml_path = self.__create_folder_and_content(
                            path_learn,
                            run_name,
                            patients_train,
                            patients_test,
                            ml_path
                        )
            elif type_set.lower() == 'cv':
                # Get the experiment options for the sets
                cv_info = ml['design'][type_set]
                n_splits = cv_info['nSplits']
                seed = cv_info['seed']
                
                # Get the training and testing sets
                patients_train, patients_test = cross_validation_split(
                    outcomes_table, 
                    n_splits, 
                    seed=seed
                )

                # If patients are not in a list
                if type(patients_train) != list and not hasattr((patients_train), "__len__"):
                    patients_train = [patients_train]
                    patients_test = [patients_test]
                
                for i in range(n_splits):
                    # Create a folder for each split/run
                    run_name = "test__{0:03}".format(i+1)
                    ml_path = self.__create_folder_and_content(
                        path_learn,
                        run_name,
                        patients_train[i],
                        patients_test[i],
                        ml_path
                    )
            else:
                raise ValueError("The type of test set is not recognized. Must be 'random' or 'institutions'.")

        # Make ml_path a dictionary to easily save it in json
        return {f"run{idx+1}": value for idx, value in enumerate(ml_path)}

    def generate_experiment(self):
        """
        Generate the json files containing all the options the experiment.
        The json files will then be used in machine learning.
        """
        # Generate the ml options dictionary
        path_ml_options = self.generate_learner_dict()

        # Fill the ml options dictionary
        self.path_ml_object = self.fill_learner_dict(path_ml_options)
        
        # Generate the experiment dictionary
        experiment_dict = self.create_experiment()
        
        # Saving the final experiment dictionary
        path_file = self.path_study / f'path_file_ml_paths__{self.experiment_label}.json'
        experiment_dict = posix_to_string(experiment_dict)  # Convert all paths to string
        save_json(path_file, experiment_dict)
        
        return path_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import logging
from copy import deepcopy
from typing import List, Sequence, Tuple, Union

import numpy as np
from nibabel import Nifti1Image
from scipy.ndimage import center_of_mass

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj
from ..utils.imref import imref3d, intrinsicToWorld, worldToIntrinsic
from ..utils.inpolygon import inpolygon
from ..utils.interp3 import interp3
from ..utils.mode import mode
from ..utils.parse_contour_string import parse_contour_string
from ..utils.strfind import strfind

_logger = logging.getLogger(__name__)


def get_roi_from_indexes(
        medscan: MEDscan, 
        name_roi: str, 
        box_string: str
    ) -> Tuple[image_volume_obj, image_volume_obj]:
    """Extracts the ROI box (+ smallest box containing the region of interest)
    and associated mask from the indexes saved in ``medscan`` scan.
    
    Args:
        medscan (MEDscan): The MEDscan class object.
        name_roi (str): name of the ROI since the a volume can have multiple
            ROIs.
        box_string (str): Specifies the size if the box containing the ROI

            - 'full': Full imaging data as output.
            - 'box': computes the smallest bounding box.
            - Ex: 'box10': 10 voxels in all three dimensions are added to \
                the smallest bounding box. The number after 'box' defines the \
                number of voxels to add.
            - Ex: '2box': Computes the smallest box and outputs double its \
                size. The number before 'box' defines the multiplication in \
                size.

    Returns:
        2-element tuple containing

        - ndarray: vol_obj, 3D array of imaging data defining the smallest box \
            containing the region of interest.
        - ndarray: roi_obj, 3D array of 1's and 0's defining the ROI in ROIbox.
    """
    # This takes care of the "Volume resection" step
    # as well using the argument "box". No fourth
    # argument means 'interp' by default.

    # PARSING OF ARGUMENTS
    try:
        name_structure_set = []
        delimiters = ["\+", "\-"]
        n_contour_data = len(medscan.data.ROI.indexes)

        name_roi, vect_plus_minus = get_sep_roi_names(name_roi, delimiters)
        contour_number = np.zeros(len(name_roi))

        if name_structure_set is None:
            name_structure_set = []

        if name_structure_set:
            name_structure_set, _ = get_sep_roi_names(name_structure_set, delimiters)
            if len(name_roi) != len(name_structure_set):
                raise ValueError(
                    "The numbers of defined ROI names and Structure Set names are not the same")

        for i in range(0, len(name_roi)):
            for j in range(0, n_contour_data):
                name_temp = medscan.data.ROI.get_roi_name(key=j)
                if name_temp == name_roi[i]:
                    if name_structure_set:
                        # FOR DICOM + RTSTRUCT
                        name_set_temp = medscan.data.ROI.get_name_set(key=j)
                        if name_set_temp == name_structure_set[i]:
                            contour_number[i] = j
                            break
                    else:
                        contour_number[i] = j
                        break

        n_roi = np.size(contour_number)
        # contour_string IS FOR EXAMPLE '3' or '1-3+2'
        contour_string = str(contour_number[0].astype(int))

        for i in range(1, n_roi):
            if vect_plus_minus[i-1] == 1:
                sign = '+'
            elif vect_plus_minus[i-1] == -1:
                sign = '-'
            contour_string = contour_string + sign + \
                str(contour_number[i].astype(int))

        if not (box_string == "full" or "box" in box_string):
            raise ValueError(
                "The third argument must either be \"full\" or contain the word \"box\".")

        contour_number, operations = parse_contour_string(contour_string)

        # INTIALIZATIONS
        if type(contour_number) is int:
            n_contour = 1
            contour_number = [contour_number]
        else:
            n_contour = len(contour_number)

        # Note: sData is a nested dictionary not an object
        spatial_ref = medscan.data.volume.spatialRef
        vol = medscan.data.volume.array.astype(np.float32)

        # APPLYING OPERATIONS ON ALL MASKS
        roi = medscan.data.get_indexes_by_roi_name(name_roi[0])
        for c in np.arange(start=1, stop=n_contour):
            if operations[c-1] == "+":
                roi += medscan.data.get_indexes_by_roi_name(name_roi[c])
            elif operations[c-1] == "-":
                roi -= medscan.data.get_indexes_by_roi_name(name_roi[c])
            else:
                raise ValueError("Unknown operation on ROI.")

            roi[roi >= 1.0] = 1.0
            roi[roi < 1.0] = 0.0

        # COMPUTING THE BOUNDING BOX
        vol, roi, new_spatial_ref = compute_box(vol=vol, roi=roi,
                                            spatial_ref=spatial_ref,
                                            box_string=box_string)

        # ARRANGE OUTPUT
        vol_obj = image_volume_obj(data=vol, spatial_ref=new_spatial_ref)
        roi_obj = image_volume_obj(data=roi, spatial_ref=new_spatial_ref)

    except Exception as e:
        message = f"\n PROBLEM WITH PRE-PROCESSING OF FEATURES IN get_roi_from_indexes():\n {e}"
        logging.error(message)
        print(message)

        if medscan:
            medscan.radiomics.image.update(
                {('scale'+(str(medscan.params.process.scale_non_text[0])).replace('.', 'dot')): 'ERROR_PROCESSING'})

    return vol_obj, roi_obj

def get_sep_roi_names(name_roi_in: str,
                      delimiters: List) -> Tuple[List[int],
                                                 np.ndarray]:
    """Seperated ROI names present in the given ROI name. An ROI name can
    have multiple ROI names seperated with curly brackets and delimeters.
    Note:
        Works only for delimiters "+" and "-".
    Args:
        name_roi_in (str): Name of ROIs that will be extracted from the imagign volume. \
                           Separated with curly brackets and delimeters. Ex: '{ED}+{ET}'.
        delimiters (List): List of delimeters of "+" and "-".
    Returns:
        2-element tuple containing 
        
        - List[int]: List of ROI names seperated and excluding curly brackets.
        - ndarray: array of 1's and -1's that defines the regions that will \
                 included and/or excluded in/from the imaging data.
    Examples:
        >>> get_sep_roi_names('{ED}+{ET}', ['+', '-'])
        ['ED', 'ET'], [1]
        >>> get_sep_roi_names('{ED}-{ET}', ['+', '-'])
        ['ED', 'ET'], [-1]
    """
    # EX:
    #name_roi_in = '{GTV-1}'
    #delimiters = ['\\+','\\-']

    # FINDING "+" and "-"
    ind_plus = strfind(string=name_roi_in, pattern=delimiters[0])
    vect_plus = np.ones(len(ind_plus))
    ind_minus = strfind(string=name_roi_in, pattern=delimiters[1])
    vect_minus = np.ones(len(ind_minus)) * -1
    ind = np.argsort(np.hstack((ind_plus, ind_minus)))
    vect_plus_minus = np.hstack((vect_plus, vect_minus))[ind]
    ind = np.hstack((ind_plus, ind_minus))[ind].astype(int)
    n_delim = np.size(vect_plus_minus)

    # MAKING SURE "+" and "-" ARE NOT INSIDE A ROIname
    ind_start = strfind(string=name_roi_in, pattern="{")
    n_roi = len(ind_start)
    ind_stop = strfind(string=name_roi_in, pattern="}")
    ind_keep = np.ones(n_delim, dtype=bool)
    for d in np.arange(n_delim):
        for r in np.arange(n_roi):
             # Thus not indise a ROI name
            if (ind_stop[r] - ind[d]) > 0 and (ind[d] - ind_start[r]) > 0:
                ind_keep[d] = False
                break

    ind = ind[ind_keep]
    vect_plus_minus = vect_plus_minus[ind_keep]

    # PARSING ROI NAMES
    if ind.size == 0:
        # Excluding the "{" and "}" at the start and end of the ROIname
        name_roi_out = [name_roi_in[1:-1]]
    else:
        n_ind = len(ind)
        # Excluding the "{" and "}" at the start and end of the ROIname
        name_roi_out = [name_roi_in[1:(ind[0]-1)]]
        for i in np.arange(start=1, stop=n_ind):
            # Excluding the "{" and "}" at the start and end of the ROIname
            name_roi_out += [name_roi_in[(ind[i-1]+2):(ind[i]-1)]]
        name_roi_out += [name_roi_in[(ind[-1]+2):-1]]

    return name_roi_out, vect_plus_minus
    
def roi_extract(vol: np.ndarray,
                roi: np.ndarray) -> np.ndarray:
    """Replaces volume intensities outside the ROI with NaN.

    Args:
        vol (ndarray): Imaging data.
        roi (ndarray): ROI mask with values of 0's and 1's.

    Returns:
        ndarray: Imaging data with original intensities in the ROI \
            and NaN for intensities outside the ROI.
    """

    vol_re = deepcopy(vol)
    vol_re[roi == 0] = np.nan

    return vol_re

def get_polygon_mask(roi_xyz: np.ndarray,
                     spatial_ref: imref3d) -> np.ndarray:
    """Computes the indexes of the ROI (Region of interest) enclosing box
    in all dimensions.

    Args:
        roi_xyz (ndarray): array of (x,y,z) triplets defining a contour in the
                           Patient-Based Coordinate System extracted from DICOM RTstruct.
        spatial_ref (imref3d): imref3d object (same functionality of MATLAB imref3d class).

    Returns:
        ndarray: 3D array of 1's and 0's defining the ROI mask.
    """

    # COMPUTING MASK
    s_z = spatial_ref.ImageSize.copy()
    roi_mask = np.zeros(s_z)
    # X,Y,Z in intrinsic image coordinates
    X, Y, Z = worldToIntrinsic(R=spatial_ref,
                               xWorld=roi_xyz[:, 0],
                               yWorld=roi_xyz[:, 1],
                               zWorld=roi_xyz[:, 2])

    points = np.transpose(np.vstack((X, Y, Z)))

    K = np.round(points[:, 2])  # Must assign the points to one slice
    closed_contours = np.unique(roi_xyz[:, 3])
    x_q = np.arange(s_z[0])
    y_q = np.arange(s_z[1])
    x_q, y_q = np.meshgrid(x_q, y_q)

    for c_c in np.arange(len(closed_contours)):
        ind = roi_xyz[:, 3] == closed_contours[c_c]
        # Taking the mode, just in case. But normally, numel(unique(K(ind)))
        # should evaluate to 1, as closed contours are meant to be defined on
        # a given slice
        select_slice = mode(K[ind]).astype(int)
        inpoly = inpolygon(x_q=x_q, y_q=y_q, x_v=points[ind, 0], y_v=points[ind, 1])
        roi_mask[:, :, select_slice] = np.logical_or(
            roi_mask[:, :, select_slice], inpoly)

    return roi_mask

def voxel_to_spatial(affine: np.ndarray,
                     voxel_pos: list) -> np.array:
    """Convert voxel position into spatial position.

    Args:
        affine (ndarray): Affine matrix.
        voxel_pos (list): A list that correspond to the location in voxel.

    Returns:
        ndarray: A numpy array that correspond to the spatial position in mm.
    """
    m = affine[:3, :3]
    translation = affine[:3, 3]
    return m.dot(voxel_pos) + translation

def spatial_to_voxel(affine: np.ndarray,
                     spatial_pos: list) -> np.array:
    """Convert spatial position into voxel position

    Args:
        affine (ndarray): Affine matrix.
        spatial_pos (list): A list that correspond to the spatial location in mm.

    Returns:
        ndarray: A numpy array that correspond to the position in the voxel.
    """
    affine = np.linalg.inv(affine)
    m = affine[:3, :3]
    translation = affine[:3, 3]
    return m.dot(spatial_pos) + translation

def crop_nifti_box(image: Nifti1Image,
                   roi: Nifti1Image,
                   crop_shape: List[int],
                   center: Union[Sequence[int], None] = None) -> Tuple[Nifti1Image,
                                                                       Nifti1Image]:
    """Crops the Nifti image and ROI.

    Args:
        image (Nifti1Image): Class for the file NIfTI1 format image that will be cropped.
        roi (Nifti1Image): Class for the file NIfTI1 format ROI that will be cropped.
        crop_shape (List[int]): The dimension of the region to crop in term of number of voxel.
        center (Union[Sequence[int], None]): A list that indicate the center of the cropping box
                                             in term of spatial position.

    Returns:
        Tuple[Nifti1Image, Nifti1Image] : Two Nifti images of the cropped image and roi
    """
    assert np.sum(np.array(crop_shape) % 2) == 0, "All elements of crop_shape should be even number."

    image_data = image.get_fdata()
    roi_data = roi.get_fdata()

    radius = [int(x / 2) - 1 for x in crop_shape]
    if center is None:
        center = list(np.array(list(center_of_mass(roi_data))).astype(int))

    center_min = np.floor(center).astype(int)
    center_max = np.ceil(center).astype(int)

    # If center_max and center_min are equal we add 1 to center_max to avoid trouble with crop.
    for i in range(3):
        center_max[i] += 1 if center_max[i] == center_min[i] else 0

    img_shape = image.header['dim'][1:4]

    # Pad the image and the ROI if its necessary
    padding = []
    for rad, cent_min, cent_max, shape in zip(radius, center_min, center_max, img_shape):
        padding.append(
            [abs(min(cent_min - rad, 0)), max(cent_max + rad + 1 - shape, 0)]
        )

    image_data = np.pad(image_data, tuple([tuple(x) for x in padding]))
    roi_data = np.pad(roi_data, tuple([tuple(x) for x in padding]))

    center_min = [center_min[i] + padding[i][0] for i in range(3)]
    center_max = [center_max[i] + padding[i][0] for i in range(3)]

    # Crop the image
    image_data = image_data[center_min[0] - radius[0]:center_max[0] + radius[0] + 1,
                center_min[1] - radius[1]:center_max[1] + radius[1] + 1,
                center_min[2] - radius[2]:center_max[2] + radius[2] + 1]
    roi_data = roi_data[center_min[0] - radius[0]:center_max[0] + radius[0] + 1,
                center_min[1] - radius[1]:center_max[1] + radius[1] + 1,
                center_min[2] - radius[2]:center_max[2] + radius[2] + 1]

    # Update the image and the ROI
    image = Nifti1Image(image_data, affine=image.affine, header=image.header)
    roi = Nifti1Image(roi_data, affine=roi.affine, header=roi.header)

    return image, roi

def crop_box(image_data: np.ndarray,
             roi_data: np.ndarray,
             crop_shape: List[int],
             center: Union[Sequence[int], None] = None) -> Tuple[np.ndarray, np.ndarray]:
    """Crops the imaging data and the ROI mask.

    Args:
        image_data (ndarray): Imaging data that will be cropped.
        roi_data (ndarray): Mask data that will be cropped.
        crop_shape (List[int]): The dimension of the region to crop in term of number of voxel.
        center (Union[Sequence[int], None]): A list that indicate the center of the cropping box 
                                             in term of spatial position.

    Returns:
        Tuple[ndarray, ndarray] : Two numpy arrays of the cropped image and roi
    """
    assert np.sum(np.array(crop_shape) % 2) == 0, "All elements of crop_shape should be even number."

    radius = [int(x / 2) - 1 for x in crop_shape]
    if center is None:
        center = list(np.array(list(center_of_mass(roi_data))).astype(int))

    center_min = np.floor(center).astype(int)
    center_max = np.ceil(center).astype(int)

    # If center_max and center_min are equal we add 1 to center_max to avoid trouble with crop.
    for i in range(3):
        center_max[i] += 1 if center_max[i] == center_min[i] else 0

    img_shape = image_data.shape

    # Pad the image and the ROI if its necessary
    padding = []
    for rad, cent_min, cent_max, shape in zip(radius, center_min, center_max, img_shape):
        padding.append(
            [abs(min(cent_min - rad, 0)), max(cent_max + rad + 1 - shape, 0)]
        )

    image_data = np.pad(image_data, tuple([tuple(x) for x in padding]))
    roi_data = np.pad(roi_data, tuple([tuple(x) for x in padding]))

    center_min = [center_min[i] + padding[i][0] for i in range(3)]
    center_max = [center_max[i] + padding[i][0] for i in range(3)]

    # Crop the image
    image_data = image_data[center_min[0] - radius[0]:center_max[0] + radius[0] + 1,
                center_min[1] - radius[1]:center_max[1] + radius[1] + 1,
                center_min[2] - radius[2]:center_max[2] + radius[2] + 1]
    roi_data = roi_data[center_min[0] - radius[0]:center_max[0] + radius[0] + 1,
                center_min[1] - radius[1]:center_max[1] + radius[1] + 1,
                center_min[2] - radius[2]:center_max[2] + radius[2] + 1]
    
    return image_data, roi_data

def compute_box(vol: np.ndarray,
                roi: np.ndarray,
                spatial_ref: imref3d,
                box_string: str) -> Tuple[np.ndarray,
                                          np.ndarray,
                                          imref3d]:
    """Computes a new box around the ROI (Region of interest) from the original box
    and updates the volume and the ``spatial_ref``.

    Args:
        vol (ndarray): ROI mask with values of 0 and 1.
        roi (ndarray): ROI mask with values of 0 and 1.
        spatial_ref (imref3d): imref3d object (same functionality of MATLAB imref3d class).
        box_string (str): Specifies the new box to be computed

            * 'full': full imaging data as output.
            * 'box': computes the smallest bounding box.
            * Ex: 'box10' means 10 voxels in all three dimensions are added to the smallest bounding box. The number \
                after 'box' defines the number of voxels to add.
            * Ex: '2box' computes the smallest box and outputs double its \
                size. The number before 'box' defines the multiplication in size.
    
    Returns: 
        3-element tuple containing

        - ndarray: 3D array of imaging data defining the smallest box containing the ROI.
        - ndarray: 3D array of 1's and 0's defining the ROI in ROIbox.
        - imref3d: The associated imref3d object imaging data.

    Todo:
        * I would not recommend parsing different settings into a string. \
            Provide two or more parameters instead, and use None if one or more \
            are not used.
        * There is no else statement, so "new_spatial_ref" might be unset
    """

    if "box" in box_string:
        comp = box_string == "box"
        box_bound = compute_bounding_box(mask=roi)
        if not comp:
            # Always returns the first appearance
            ind_box = box_string.find("box")
            # Addition of a certain number of voxels in all dimensions
            if ind_box == 0:
                n_v = float(box_string[(ind_box+3):])
                n_v = np.array([n_v, n_v, n_v]).astype(int)
            else:  # Multiplication of the size of the box
                factor = float(box_string[0:ind_box])
                size_box = np.diff(box_bound, axis=1) + 1
                new_box = size_box * factor
                n_v = np.round((new_box - size_box)/2.0).astype(int)

            o_k = False

            while not o_k:
                border = np.zeros([3, 2])
                border[0, 0] = box_bound[0, 0] - n_v[0]
                border[0, 1] = box_bound[0, 1] + n_v[0]
                border[1, 0] = box_bound[1, 0] - n_v[1]
                border[1, 1] = box_bound[1, 1] + n_v[1]
                border[2, 0] = box_bound[2, 0] - n_v[2]
                border[2, 1] = box_bound[2, 1] + n_v[2]
                border = border + 1
                check1 = np.sum(border[:, 0] > 0)
                check2 = border[0, 1] <= vol.shape[0]
                check3 = border[1, 1] <= vol.shape[1]
                check4 = border[2, 1] <= vol.shape[2]

                check = check1 + check2 + check3 + check4

                if check == 6:
                    o_k = True
                else:
                    n_v = np.floor(n_v / 2.0)
                    if np.sum(n_v) == 0.0:
                        o_k = True
                        n_v = [0.0, 0.0, 0.0]
        else:
            # Will compute the smallest bounding box possible
            n_v = [0.0, 0.0, 0.0]

        box_bound[0, 0] -= n_v[0]
        box_bound[0, 1] += n_v[0]
        box_bound[1, 0] -= n_v[1]
        box_bound[1, 1] += n_v[1]
        box_bound[2, 0] -= n_v[2]
        box_bound[2, 1] += n_v[2]

        box_bound = box_bound.astype(int)

        vol = vol[box_bound[0, 0]:box_bound[0, 1] + 1,
                  box_bound[1, 0]:box_bound[1, 1] + 1,
                  box_bound[2, 0]:box_bound[2, 1] + 1]
        roi = roi[box_bound[0, 0]:box_bound[0, 1] + 1,
                  box_bound[1, 0]:box_bound[1, 1] + 1,
                  box_bound[2, 0]:box_bound[2, 1] + 1]

        # Resolution in mm, nothing has changed here in terms of resolution;
        # XYZ format here.
        res = np.array([spatial_ref.PixelExtentInWorldX,
                        spatial_ref.PixelExtentInWorldY,
                        spatial_ref.PixelExtentInWorldZ])

        # IJK, as required by imref3d
        size_box = (np.diff(box_bound, axis=1) + 1).tolist()
        size_box[0] = size_box[0][0]
        size_box[1] = size_box[1][0]
        size_box[2] = size_box[2][0]
        x_limit, y_limit, z_limit = intrinsicToWorld(spatial_ref, 
                                                box_bound[0, 0],
                                                box_bound[1, 0],
                                                box_bound[2, 0])
        new_spatial_ref = imref3d(size_box, res[0], res[1], res[2])

        # The limit is defined as the border of the first pixel
        new_spatial_ref.XWorldLimits = new_spatial_ref.XWorldLimits - (
            new_spatial_ref.XWorldLimits[0] - (x_limit - res[0]/2))
        new_spatial_ref.YWorldLimits = new_spatial_ref.YWorldLimits - (
            new_spatial_ref.YWorldLimits[0] - (y_limit - res[1]/2))
        new_spatial_ref.ZWorldLimits = new_spatial_ref.ZWorldLimits - (
            new_spatial_ref.ZWorldLimits[0] - (z_limit - res[2]/2))

    elif "full" in box_string:
        new_spatial_ref = spatial_ref

    return vol, roi, new_spatial_ref

def compute_bounding_box(mask:np.ndarray) -> np.ndarray:
    """Computes the indexes of the ROI (Region of interest) enclosing box 
    in all dimensions.

    Args:
        mask (ndarray): ROI mask with values of 0 and 1.

    Returns:
        ndarray: An array containing the indexes of the bounding box.
    """

    indices = np.where(np.reshape(mask, np.size(mask), order='F') == 1)
    iv, jv, kv = np.unravel_index(indices, np.shape(mask), order='F')
    box_bound = np.zeros((3, 2))
    box_bound[0, 0] = np.min(iv)
    box_bound[0, 1] = np.max(iv)
    box_bound[1, 0] = np.min(jv)
    box_bound[1, 1] = np.max(jv)
    box_bound[2, 0] = np.min(kv)
    box_bound[2, 1] = np.max(kv)

    return box_bound.astype(int)

def get_roi(medscan: MEDscan,
            name_roi: str,
            box_string: str,
            interp=False) -> Union[image_volume_obj,
                                   image_volume_obj]:
    """Computes the ROI box (box containing the region of interest)
    and associated mask from MEDscan object.

    Args:
        medscan (MEDscan): The MEDscan class object.
        name_roi (str): name of the ROI since the a volume can have multuiple ROIs.
        box_string (str): Specifies the size if the box containing the ROI

                          - 'full': full imaging data as output.
                          - 'box': computes the smallest bounding box.
                          - Ex: 'box10': 10 voxels in all three dimensions are added to \
                            the smallest bounding box. The number after 'box' defines the \
                            number of voxels to add.
                          - Ex: '2box': Computes the smallest box and outputs double its \
                            size. The number before 'box' defines the multiplication in size.

        interp (bool): True if we need to use an interpolation for box computation.

    Returns:
        2-element tuple containing

        - image_volume_obj: 3D array of imaging data defining box containing the ROI. \
            vol.data is the 3D array, vol.spatialRef is its associated imref3d object.
        - image_volume_obj: 3D array of 1's and 0's defining the ROI. \
            roi.data is the 3D array, roi.spatialRef is its associated imref3d object.
    """
    # PARSING OF ARGUMENTS
    try:
        name_structure_set = []
        delimiters = ["\+", "\-"]
        n_contour_data = len(medscan.data.ROI.indexes)

        name_roi, vect_plus_minus = get_sep_roi_names(name_roi, delimiters)
        contour_number = np.zeros(len(name_roi))

        if name_structure_set is None:
            name_structure_set = []

        if name_structure_set:
            name_structure_set, _ = get_sep_roi_names(name_structure_set, delimiters)
            if len(name_roi) != len(name_structure_set):
                raise ValueError(
                    "The numbers of defined ROI names and Structure Set names are not the same")

        for i in range(0, len(name_roi)):
            for j in range(0, n_contour_data):
                name_temp = medscan.data.ROI.get_roi_name(key=j)
                if name_temp == name_roi[i]:
                    if name_structure_set:
                        # FOR DICOM + RTSTRUCT
                        name_set_temp = medscan.data.ROI.get_name_set(key=j)
                        if name_set_temp == name_structure_set[i]:
                            contour_number[i] = j
                            break
                    else:
                        contour_number[i] = j
                        break

        n_roi = np.size(contour_number)
        # contour_string IS FOR EXAMPLE '3' or '1-3+2'
        contour_string = str(contour_number[0].astype(int))

        for i in range(1, n_roi):
            if vect_plus_minus[i-1] == 1:
                sign = '+'
            elif vect_plus_minus[i-1] == -1:
                sign = '-'
            contour_string = contour_string + sign + \
                str(contour_number[i].astype(int))

        if not (box_string == "full" or "box" in box_string):
            raise ValueError(
                "The third argument must either be \"full\" or contain the word \"box\".")

        if type(interp) != bool:
            raise ValueError(
                "If present (i.e. it is optional), the fourth argument must be bool")

        contour_number, operations = parse_contour_string(contour_string)

        # INTIALIZATIONS
        if type(contour_number) is int:
            n_contour = 1
            contour_number = [contour_number]
        else:
            n_contour = len(contour_number)

        roi_mask_list = []
        if medscan.type not in ["PTscan", "CTscan", "MRscan", "ADCscan"]:
            raise ValueError("Unknown scan type.")

        spatial_ref = medscan.data.volume.spatialRef
        vol = medscan.data.volume.array.astype(np.float32)

        # COMPUTING ALL MASKS
        for c in np.arange(start=0, stop=n_contour):
            contour = contour_number[c]
            # GETTING THE XYZ POINTS FROM medscan
            roi_xyz = medscan.data.ROI.get_indexes(key=contour).copy()

            # APPLYING ROTATION TO XYZ POINTS (if necessary --> MRscan)
            if hasattr(medscan.data.volume, 'scan_rot') and medscan.data.volume.scan_rot is not None:
                roi_xyz[:, [0, 1, 2]] = np.transpose(
                    medscan.data.volume.scan_rot @ np.transpose(roi_xyz[:, [0, 1, 2]]))

            # APPLYING TRANSLATION IF SIMULATION STRUCTURE AS INPUT
            # (software STAMP utility)
            if hasattr(medscan.data.volume, 'transScanToModel'):
                translation = medscan.data.volume.transScanToModel
                roi_xyz[:, 0] += translation[0]
                roi_xyz[:, 1] += translation[1]
                roi_xyz[:, 2] += translation[2]

            # COMPUTING THE ROI MASK
            # Problem here in compute_roi.m: If the volume is a full-body CT and the
            # slice interpolation process occurs, a lot of RAM will be used.
            # One solution could be to a priori compute the bounding box before
            # computing the ROI (using XYZ points). But we still want the user to
            # be able to fully use the "box" argument, so we are fourré...TO SOLVE!
            roi_mask_list += [compute_roi(roi_xyz=roi_xyz, 
                                        spatial_ref=spatial_ref,
                                        orientation=medscan.data.orientation,
                                        scan_type=medscan.type,
                                        interp=interp).astype(np.float32)]

        # APPLYING OPERATIONS ON ALL MASKS
        roi = roi_mask_list[0]
        for c in np.arange(start=1, stop=n_contour):
            if operations[c-1] == "+":
                roi += roi_mask_list[c]
            elif operations[c-1] == "-":
                roi -= roi_mask_list[c]
            else:
                raise ValueError("Unknown operation on ROI.")

            roi[roi >= 1.0] = 1.0
            roi[roi < 1.0] = 0.0

        # COMPUTING THE BOUNDING BOX
        vol, roi, new_spatial_ref = compute_box(vol=vol, 
                                            roi=roi,
                                            spatial_ref=spatial_ref,
                                            box_string=box_string)

        # ARRANGE OUTPUT
        vol_obj = image_volume_obj(data=vol, spatial_ref=new_spatial_ref)
        roi_obj = image_volume_obj(data=roi, spatial_ref=new_spatial_ref)

    except Exception as e:
        message = f"\n PROBLEM WITH PRE-PROCESSING OF FEATURES IN get_roi(): \n {e}"
        _logger.error(message)
        print(message)

        if medscan:
            medscan.radiomics.image.update(
                {('scale'+(str(medscan.params.process.scale_non_text[0])).replace('.', 'dot')): 'ERROR_PROCESSING'})

    return vol_obj, roi_obj

def compute_roi(roi_xyz: np.ndarray,
                spatial_ref: imref3d,
                orientation: str,
                scan_type: str,
                interp=False) -> np.ndarray:
    """Computes the ROI (Region of interest) mask using the XYZ coordinates.

    Args:
        roi_xyz (ndarray): array of (x,y,z) triplets defining a contour in the Patient-Based
                           Coordinate System extracted from DICOM RTstruct.
        spatial_ref (imref3d): imref3d object (same functionality of MATLAB imref3d class).
        orientation (str): Imaging data ``orientation`` (axial, sagittal or coronal).
        scan_type (str): Imaging modality (MRscan, CTscan...).
        interp (bool): Specifies if we need to use an interpolation \
            process prior to :func:`get_polygon_mask()` in the slice axis direction.

            - True: Interpolation is performed in the slice axis dimensions. To be further \
                tested, thus please use with caution (True is safer).
            - False (default): No interpolation. This can definitely be safe \
                when the RTstruct has been saved specifically for the volume of \
                interest.
        
    Returns:
        ndarray: 3D array of 1's and 0's defining the ROI mask.

    Todo:
        - Using interpolation: this part needs to be further tested.
        - Consider changing to 'if statement'. Changing ``interp`` variable here will change the ``interp`` variable everywhere
    """

    while interp:
        # Initialization
        if orientation == "Axial":
            dim_ijk = 2
            dim_xyz = 2
            direction = "Z"
            # Only the resolution in 'Z' will be changed
            res_xyz = np.array([spatial_ref.PixelExtentInWorldX,
                               spatial_ref.PixelExtentInWorldY, 0.0])
        elif orientation == "Sagittal":
            dim_ijk = 0
            dim_xyz = 1
            direction = "Y"
            # Only the resolution in 'Y' will be changed
            res_xyz = np.array([spatial_ref.PixelExtentInWorldX, 0.0,
                               spatial_ref.PixelExtentInWorldZ])
        elif orientation == "Coronal":
            dim_ijk = 1
            dim_xyz = 0
            direction = "X"
            # Only the resolution in 'X' will be changed
            res_xyz = np.array([0.0, spatial_ref.PixelExtentInWorldY,
                               spatial_ref.PixelExtentInWorldZ])
        else:
            raise ValueError(
                "Provided orientation is not one of \"Axial\", \"Sagittal\", \"Coronal\".")

        # Creating new imref3d object for sample points (with slice dimension
        # similar to original volume
        # where RTstruct was created)
        # Slice spacing in mm
        slice_spacing = find_spacing(
            roi_xyz[:, dim_ijk], scan_type).astype(np.float32)

        # Only one slice found in the function "find_spacing" on the above line.
        # We thus must set "slice_spacing" to the slice spacing of the queried
        # volume, and no interpolation will be performed.
        if slice_spacing is None:
            slice_spacing = spatial_ref.PixelExtendInWorld(axis=direction)

        new_size = round(spatial_ref.ImageExtentInWorld(
            axis=direction) / slice_spacing)
        res_xyz[dim_xyz] = slice_spacing
        s_z = spatial_ref.ImageSize.copy()
        s_z[dim_ijk] = new_size

        xWorldLimits = spatial_ref.XWorldLimits.copy()
        yWorldLimits = spatial_ref.YWorldLimits.copy()
        zWorldLimits = spatial_ref.ZWorldLimits.copy()

        new_spatial_ref = imref3d(imageSize=s_z, 
                                pixelExtentInWorldX=res_xyz[0],
                                pixelExtentInWorldY=res_xyz[1],
                                pixelExtentInWorldZ=res_xyz[2],
                                xWorldLimits=xWorldLimits,
                                yWorldLimits=yWorldLimits,
                                zWorldLimits=zWorldLimits)

        diff = (new_spatial_ref.ImageExtentInWorld(axis=direction) -
                spatial_ref.ImageExtentInWorld(axis=direction))

        if np.abs(diff) >= 0.01:
            # Sampled and queried volume are considered "different".
            new_limit = spatial_ref.WorldLimits(axis=direction)[0] - diff / 2.0

            # Sampled volume is now centered on queried volume.
            new_spatial_ref.WorldLimits(axis=direction, newValue=(new_spatial_ref.WorldLimits(axis=direction) -
                                                                (new_spatial_ref.WorldLimits(axis=direction)[0] - 
                                                                 new_limit)))
        else:
            # Less than a 0.01 mm, sampled and queried volume are considered
            # to be the same. At this point,
            # spatial_ref and new_spatial_ref may have differed due to data
            # manipulation, so we simply compute
            # the ROI mask with spatial_ref (i.e. simply using "poly2mask.m"),
            # without performing interpolation.
            interp = False
            break  # Getting out of the "while" statement

        V = get_polygon_mask(roi_xyz, new_spatial_ref)

        # Getting query points (x_q,y_q,z_q) of output roi_mask
        sz_q = spatial_ref.ImageSize
        x_qi = np.arange(sz_q[0])
        y_qi = np.arange(sz_q[1])
        z_qi = np.arange(sz_q[2])
        x_qi, y_qi, z_qi = np.meshgrid(x_qi, y_qi, z_qi, indexing='ij')

        # Getting queried mask
        v_q = interp3(V=V, x_q=x_qi, y_q=y_qi, z_q=z_qi, method="cubic")
        roi_mask = v_q
        roi_mask[v_q < 0.5] = 0
        roi_mask[v_q >= 0.5] = 1

        # Getting out of the "while" statement
        interp = False

    # SIMPLY USING "poly2mask.m" or "inpolygon.m". "inpolygon.m" is slower, but
    # apparently more accurate.
    if not interp:
        # Using the inpolygon.m function. To be further tested.
        roi_mask = get_polygon_mask(roi_xyz, spatial_ref)

    return roi_mask

def find_spacing(points: np.ndarray,
                 scan_type: str) -> float:
    """Finds the slice spacing in mm.

    Note:
        This function works for points from at least 2 slices. If only
        one slice is present, the function returns a None.

    Args:
        points (ndarray): Array of (x,y,z) triplets defining a contour in the
            Patient-Based Coordinate System extracted from DICOM RTstruct.
        scan_type (str): Imaging modality (MRscan, CTscan...) 

    Returns:
        float: Slice spacing in mm.
    """
    decim_keep = 4  # We keep at most 4 decimals to find the slice spacing.

    # Rounding to the nearest 0.1 mm, MRI is more problematic due to arbitrary
    # orientations allowed for imaging volumes.
    if scan_type == "MRscan":
        slices = np.unique(np.around(points, 1))
    else:
        slices = np.unique(np.around(points, 2))

    n_slices = len(slices)
    if n_slices == 1:
        return None

    diff = np.abs(np.diff(slices))
    diff = np.round(diff, decim_keep)
    slice_spacing, nOcc = mode(x=diff, return_counts=True)
    if np.max(nOcc) == 1:
        slice_spacing = np.mean(diff)

    return slice_spacing
